# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Base class for templates.
Contains common code used by multiple templates.
"""
import functools
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Optional

import pandas as pd

from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.time_properties_forecast import get_forecast_time_properties
from greykite.framework.pipeline.utils import get_basic_pipeline
from greykite.framework.templates.autogen.forecast_config import ForecastConfig
from greykite.framework.templates.forecast_config_defaults import ForecastConfigDefaults
from greykite.framework.templates.template_interface import TemplateInterface
from greykite.sklearn.estimator.base_forecast_estimator import BaseForecastEstimator


class BaseTemplate(TemplateInterface, ForecastConfigDefaults, ABC):
    """Base template with common code used by multiple templates.

    Provides a particular modular approach to implement
    `~greykite.framework.templates.template_interface.TemplateInterface.apply_template_for_pipeline_params`.

    Includes the config defaults from
    `~greykite.framework.templates.forecast_config_defaults.ForecastConfigDefaults`.

    Subclasses must provide these properties / functions used by ``apply_template_for_pipeline_params``:

        - estimator (__init__ default value)
        - get_regressor_cols
        - get_hyperparameter_grid

    Subclasses may optionally want to override:

        - get_pipeline
        - get_forecast_time_properties
        - apply_metadata_defaults
        - apply_evaluation_metric_defaults
        - apply_evaluation_period_defaults
        - apply_computation_defaults
        - apply_model_components_defaults
        - apply_forecast_config_defaults
    """
    def __init__(self, estimator: BaseForecastEstimator):
        # See attributes of `TemplateInterface` and `ForecastConfigDefaults`.
        # Note that `self.config` includes modifications after applying default values.
        super().__init__()

        self._estimator: BaseForecastEstimator = estimator
        """The estimator instance to use as the final step in the pipeline.
        An instance of `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
        """

        # Attributes used by `~greykite.framework.templates.base_template.apply_template_for_pipeline_params`.
        self.score_func = None
        """Score function used to select optimal model in CV."""
        self.score_func_greater_is_better = None
        """True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        """
        self.regressor_cols = None
        """A list of regressor columns used in the training and prediction DataFrames.
        If None, no regressor columns are used.
        """
        self.pipeline = None
        """Pipeline to fit. The final named step must be called "estimator"."""
        self.time_properties = None
        """Time properties dictionary (likely produced by
        `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
        """
        self.hyperparameter_grid = None
        """Sets properties of the steps in the pipeline,
        and specifies combinations to search over.
        Should be valid input to `sklearn.model_selection.GridSearchCV` (param_grid)
        or `sklearn.model_selection.RandomizedSearchCV` (param_distributions).
        """

    @property
    def estimator(self):
        """The estimator instance to use as the final step in the pipeline.
        An instance of `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`.
        """
        return self._estimator

    @abstractmethod
    def get_regressor_cols(self):
        """Returns regressor column names.

        To be implemented by subclass.

        Available parameters:

            - self.df
            - self.config
            - self.score_func
            - self.score_func_greater_is_better

        Returns
        -------
        regressor_cols : `list` [`str`] or None
            See `~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        pass

    def get_pipeline(self):
        """Returns pipeline.

        Implementation may be overridden by subclass
        if a different pipeline is desired.

        Uses ``self.estimator``, ``self.score_func``,
        ``self.score_func_greater_is_better``, ``self.config``,
        ``self.regressor_cols``.

        Available parameters:

            - self.df
            - self.config
            - self.score_func
            - self.score_func_greater_is_better
            - self.regressor_cols
            - self.estimator

        Returns
        -------
        pipeline : `sklearn.pipeline.Pipeline`
            See `~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        return get_basic_pipeline(
            estimator=self.estimator,
            score_func=self.score_func,
            score_func_greater_is_better=self.score_func_greater_is_better,
            agg_periods=self.config.evaluation_metric_param.agg_periods,
            agg_func=self.config.evaluation_metric_param.agg_func,
            relative_error_tolerance=self.config.evaluation_metric_param.relative_error_tolerance,
            coverage=self.config.coverage,
            null_model_params=self.config.evaluation_metric_param.null_model_params,
            regressor_cols=self.regressor_cols)

    def get_forecast_time_properties(self):
        """Returns forecast time parameters.

        Uses ``self.df``, ``self.config``, ``self.regressor_cols``.

        Available parameters:

            - self.df
            - self.config
            - self.score_func
            - self.score_func_greater_is_better
            - self.regressor_cols
            - self.estimator
            - self.pipeline

        Returns
        -------
        time_properties : `dict` [`str`, `any`] or None, default None
            Time properties dictionary (likely produced by
            `~greykite.common.time_properties_forecast.get_forecast_time_properties`)
            with keys:

            ``"period"`` : `int`
                Period of each observation (i.e. minimum time between observations, in seconds).
            ``"simple_freq"`` : `SimpleTimeFrequencyEnum`
                ``SimpleTimeFrequencyEnum`` member corresponding to data frequency.
            ``"num_training_points"`` : `int`
                Number of observations for training.
            ``"num_training_days"`` : `int`
                Number of days for training.
            ``"start_year"`` : `int`
                Start year of the training period.
            ``"end_year"`` : `int`
                End year of the forecast period.
            ``"origin_for_time_vars"`` : `float`
                Continuous time representation of the first date in ``df``.
        """
        return get_forecast_time_properties(
            df=self.df,
            time_col=self.config.metadata_param.time_col,
            value_col=self.config.metadata_param.value_col,
            freq=self.config.metadata_param.freq,
            train_end_date=self.config.metadata_param.train_end_date,
            regressor_cols=self.regressor_cols,
            forecast_horizon=self.config.forecast_horizon)

    @abstractmethod
    def get_hyperparameter_grid(self):
        """Returns hyperparameter grid.

        To be implemented by subclass.

        Available parameters:

            - self.df
            - self.config
            - self.score_func
            - self.score_func_greater_is_better
            - self.regressor_cols
            - self.estimator
            - self.pipeline
            - self.time_properties

        Returns
        -------
        hyperparameter_grid : `dict`, `list` [`dict`] or None
            See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
            The output dictionary values are lists, combined in grid search.
        """
        pass

    def apply_template_decorator(func):
        """Decorator for ``apply_template_for_pipeline_params`` function.

        By default, this applies ``apply_forecast_config_defaults`` to ``config``.

        Subclass may override this for pre/post processing of
        ``apply_template_for_pipeline_params``, such as input validation.
        In this case, ``apply_template_for_pipeline_params`` must also be implemented in the subclass.
        """
        @functools.wraps(func)
        def process_wrapper(self, df: pd.DataFrame, config: Optional[ForecastConfig] = None):
            # Sets defaults and makes a copy of ``config``
            # All subclasses should keep this line.
            config = self.apply_forecast_config_defaults(config)
            # <optional processing before calling `func`, if needed>
            pipeline_params = func(self, df, config)
            # <optional postprocessing after calling `func`, if needed>
            return pipeline_params
        return process_wrapper

    @apply_template_decorator
    def apply_template_for_pipeline_params(
            self,
            df: pd.DataFrame,
            config: Optional[ForecastConfig] = None) -> Dict:
        """Implements template interface method.
        Takes input data and optional configuration parameters
        to customize the model. Returns a set of parameters to call
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.

        See template interface for parameters and return value.

        Uses the methods in this class to set:

            - ``"regressor_cols"`` : get_regressor_cols()
            - ``"pipeline"`` : get_pipeline()
            - ``"time_properties"`` : get_forecast_time_properties()
            - ``"hyperparameter_grid"`` : get_hyperparameter_grid()

        All other parameters are taken directly from ``config``.
        """
        self.df = df
        self.config = config
        # Defines score_func, score_func_greater_is_better
        # Sets `score_func` to a string instead of a function, so CV results are
        # reported as "mean_test_{short_name}" instead of "mean_test_score".
        metric = EvaluationMetricEnum[config.evaluation_metric_param.cv_selection_metric]
        self.score_func = metric.name
        self.score_func_greater_is_better = metric.get_metric_greater_is_better()

        self.regressor_cols = self.get_regressor_cols()
        self.pipeline = self.get_pipeline()
        self.time_properties = self.get_forecast_time_properties()
        self.hyperparameter_grid = self.get_hyperparameter_grid()

        self.pipeline_params = dict(
            # input
            df=self.df,
            time_col=self.config.metadata_param.time_col,
            value_col=self.config.metadata_param.value_col,
            date_format=self.config.metadata_param.date_format,
            freq=self.config.metadata_param.freq,
            train_end_date=self.config.metadata_param.train_end_date,
            anomaly_info=self.config.metadata_param.anomaly_info,
            # model
            pipeline=self.pipeline,
            regressor_cols=self.regressor_cols,
            estimator=None,  # ignored when `pipeline` is provided
            hyperparameter_grid=self.hyperparameter_grid,
            hyperparameter_budget=self.config.computation_param.hyperparameter_budget,
            n_jobs=self.config.computation_param.n_jobs,
            verbose=self.config.computation_param.verbose,
            # forecast
            forecast_horizon=self.config.forecast_horizon,
            coverage=self.config.coverage,
            test_horizon=self.config.evaluation_period_param.test_horizon,
            periods_between_train_test=self.config.evaluation_period_param.periods_between_train_test,
            agg_periods=self.config.evaluation_metric_param.agg_periods,
            agg_func=self.config.evaluation_metric_param.agg_func,
            # evaluation
            score_func=self.score_func,
            score_func_greater_is_better=self.score_func_greater_is_better,
            cv_report_metrics=self.config.evaluation_metric_param.cv_report_metrics,
            null_model_params=self.config.evaluation_metric_param.null_model_params,
            relative_error_tolerance=self.config.evaluation_metric_param.relative_error_tolerance,
            # CV
            cv_horizon=self.config.evaluation_period_param.cv_horizon,
            cv_min_train_periods=self.config.evaluation_period_param.cv_min_train_periods,
            cv_expanding_window=self.config.evaluation_period_param.cv_expanding_window,
            cv_periods_between_splits=self.config.evaluation_period_param.cv_periods_between_splits,
            cv_periods_between_train_test=self.config.evaluation_period_param.cv_periods_between_train_test,
            cv_max_splits=self.config.evaluation_period_param.cv_max_splits,
        )
        return self.pipeline_params

    # `apply_template_decorator` needs to be a static method and take `func` as the
    # only argument (not self). It also needs to be defined in this class to allow
    # override. If we use the @staticmethod decorator above, this error appears:
    #   `TypeError: 'staticmethod' object is not callable`
    # However, if we call staticmethod at the bottom of the class, after it is
    # applied to `apply_template_for_pipeline_params`, it works.
    apply_template_decorator = staticmethod(apply_template_decorator)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Sayan Patra
"""Class to load datasets in UnivariateTimeSeries format"""

from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.data_loader import DataLoader
from greykite.framework.input.univariate_time_series import UnivariateTimeSeries


class DataLoaderTS(DataLoader):
    """Returns datasets included in the library in `pandas.DataFrame` or
    `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries` format.

    Extends `~greykite.common.data_loader.DataLoader`
    """
    def __init__(self):
        super().__init__()

    def load_peyton_manning_ts(self):
        """Loads the Daily Peyton Manning dataset.

        This dataset contains log daily page views for the Wikipedia page for Peyton Manning.
        One of the primary datasets used for demonstrations by Facebook ``Prophet`` algorithm.
        Source: https://github.com/facebook/prophet/blob/master/examples/example_wp_log_peyton_manning.csv

        Below is the dataset attribute information:

            ts : date of the page view
            y : log of the number of page views

        Returns
        -------
        ts : `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries`
            Peyton Manning page views data. Time and value column:

                ``time_col`` : "ts"
                    Date of the page view.
                ``value_col`` : "y"
                    Log of the number of page views.
        """
        df = self.load_peyton_manning()
        ts = UnivariateTimeSeries()
        ts.load_data(
            df=df,
            time_col=TIME_COL,
            value_col=VALUE_COL,
            freq="1D"
        )
        return ts

    def load_parking_ts(self, system_code_number=None):
        """Loads the Hourly Parking dataset.

        This dataset contains occupancy rates (8:00 to 16:30) from 2016/10/04 to 2016/12/19
        from car parks in Birmingham that are operated by NCP from Birmingham City Council.
        Source: https://archive.ics.uci.edu/ml/datasets/Parking+Birmingham
        UK Open Government Licence (OGL)

        Below is the dataset attribute information:

            SystemCodeNumber: car park ID
            Capacity: car park capacity
            Occupancy: car park occupancy rate
            LastUpdated: date and time of the measure

        Parameters
        ----------
        system_code_number : `str` or None, default None
            If None, occupancy rate is averaged across all the ``SystemCodeNumber``.
            Else only the occupancy rate of the given ``system_code_number`` is returned.

        Returns
        -------
        ts : `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries`
            Parking data. Time and value column:

                ``time_col`` : "LastUpdated"
                    Date and Time of the Occupancy Rate, rounded to the nearest half hour.
                ``value_col`` : "OccupancyRatio"
                    ``Occupancy`` divided by ``Capacity``.
        """
        df = self.load_parking(system_code_number=system_code_number)
        ts = UnivariateTimeSeries()
        ts.load_data(
            df=df,
            time_col="LastUpdated",
            value_col="OccupancyRatio",
            freq="30min",
        )
        return ts

    def load_bikesharing_ts(self):
        """Loads the Hourly Bike Sharing Count dataset.

        This dataset contains aggregated hourly count of the number of rented bikes.
        The data also includes weather data: Maximum Daily temperature (tmax);
        Minimum Daily Temperature (tmin); Precipitation (pn)
        The raw bike-sharing data is provided by Capital Bikeshares.
        Source: https://www.capitalbikeshare.com/system-data
        The raw weather data (Baltimore-Washington INTL Airport)
        https://www.ncdc.noaa.gov/data-access/land-based-station-data

        Below is the dataset attribute information:

            ts : hour and date
            count : number of shared bikes
            tmin : minimum daily temperature
            tmax : maximum daily temperature
            pn : precipitation

        Returns
        -------
        ts : `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries`
            Bike Sharing Count data. Time and value column:

                ``time_col`` : "ts"
                    Hour and Date.
                ``value_col`` : "y"
                     Number of rented bikes across Washington DC.

            Additional regressors:

                "tmin" : minimum daily temperature
                "tmax" : maximum daily temperature
                "pn" : precipitation
        """
        df = self.load_bikesharing()
        ts = UnivariateTimeSeries()
        ts.load_data(
            df=df,
            time_col="ts",
            value_col="count",
            freq="H",
            regressor_cols=["tmin", "tmax", "pn"]
        )
        return ts

    def load_beijing_pm_ts(self):
        """Loads the Beijing Particulate Matter (PM2.5) dataset.
        https://archive.ics.uci.edu/ml/datasets/Beijing+PM2.5+Data

        This hourly data set contains the PM2.5 data of US Embassy in Beijing. Meanwhile, meteorological data
        from Beijing Capital International Airport are also included.

        The dataset's time period is between Jan 1st, 2010 to Dec 31st, 2014. Missing data are denoted as NA.

        Below is the dataset attribute information:

            No : row number
            year : year of data in this row
            month : month of data in this row
            day : day of data in this row
            hour : hour of data in this row
            pm2.5: PM2.5 concentration (ug/m^3)
            DEWP : dew point (celsius)
            TEMP : temperature (celsius)
            PRES : pressure (hPa)
            cbwd : combined wind direction
            Iws : cumulated wind speed (m/s)
            Is : cumulated hours of snow
            Ir : cumulated hours of rain

        Returns
        -------
        ts : `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries`
            Beijing PM2.5 data. Time and value column:

                ``time_col`` : TIME_COL
                    hourly timestamp
                ``value_col`` : "pm"
                     PM2.5 concentration (ug/m^3)

            Additional regressors:

                "dewp" : dew point (celsius)
                "temp" : temperature (celsius)
                "pres" : pressure (hPa)
                "cbwd" : combined wind direction
                "iws" : cumulated wind speed (m/s)
                "is" : cumulated hours of snow
                "ir" : cumulated hours of rain
        """
        df = self.load_beijing_pm()
        ts = UnivariateTimeSeries()
        ts.load_data(
            df=df,
            time_col=TIME_COL,
            value_col="pm",
            freq="H",
            regressor_cols=["dewp", "temp", "pres", "cbwd", "iws", "is", "ir"]
        )
        return ts

    def load_data_ts(self, data_name, **kwargs):
        """Loads dataset by name from the internal data library.

        Parameters
        ----------
        data_name : `str`
            Dataset to load from the internal data library.

        Returns
        -------
        ts : `~greykite.framework.input.univariate_time_series.UnivariateTimeSeries`
            Has the requested ``data_name``.
        """
        if data_name == "daily_peyton_manning":
            ts = self.load_peyton_manning_ts()
        elif data_name == "hourly_parking":
            ts = self.load_parking_ts(**kwargs)
        elif data_name == "hourly_bikesharing":
            ts = self.load_bikesharing_ts()
        elif data_name == "hourly_beijing_pm":
            ts = self.load_beijing_pm_ts()
        else:
            data_inventory = self.get_data_inventory()
            raise ValueError(f"Input data name '{data_name}' is not recognized. "
                             f"Must be one of {data_inventory}.")

        return ts



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Rachit Arora

from typing import Type


class SilverkiteColumn:
    """Silverkite feature sets for sub-daily data."""

    COLS_HOUR_OF_WEEK: str = "hour_of_week"
    """Silverkite feature_sets_enabled key. constant hour of week effect"""

    COLS_WEEKEND_SEAS: str = "is_weekend:daily_seas"
    """Silverkite feature_sets_enabled key. daily seasonality interaction with is_weekend"""

    COLS_DAY_OF_WEEK_SEAS: str = "day_of_week:daily_seas"
    """Silverkite feature_sets_enabled key. daily seasonality interaction with day of week"""

    COLS_TREND_DAILY_SEAS: str = "trend:is_weekend:daily_seas"
    """Silverkite feature_sets_enabled key. allow daily seasonality to change over time, depending on is_weekend"""

    COLS_EVENT_SEAS: str = "event:daily_seas"
    """Silverkite feature_sets_enabled key. allow sub-daily event effects"""

    # silverkite feature sets for sub-weekly data
    COLS_EVENT_WEEKEND_SEAS: str = "event:is_weekend:daily_seas"
    """Silverkite feature_sets_enabled key. allow sub-daily event effect to interact with is_weekend"""

    COLS_DAY_OF_WEEK: str = "day_of_week"
    """Silverkite feature_sets_enabled key. constant day of week effect"""

    COLS_TREND_WEEKEND: str = "trend:is_weekend"
    """Silverkite feature_sets_enabled key. allow trend (growth, changepoints) to interact with is_weekend"""

    COLS_TREND_DAY_OF_WEEK: str = "trend:day_of_week"
    """Silverkite feature_sets_enabled key. allow trend to interact with day of week"""

    COLS_TREND_WEEKLY_SEAS: str = "trend:weekly_seas"
    """Silverkite feature_sets_enabled key. allow weekly seasonality to change over time"""


class SilverkiteColumnMixin:
    """Provides a mixin for the SilverkiteColumn constants"""

    def get_silverkite_column(self) -> Type[SilverkiteColumn]:
        """Return the SilverkiteColumn constants"""
        return SilverkiteColumn



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Template interface. All templates must implement this interface to
work with `model_templates.py`.
"""
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Optional

import pandas as pd

from greykite.framework.templates.autogen.forecast_config import ForecastConfig


class TemplateInterface(ABC):
    """Defines the interface for the template class.

    Contains a single function that takes input data and forecast config,
    and returns parameters to call
    :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    """
    @abstractmethod
    def __init__(self):
        """Attributes are the parameters and return value of
        `greykite.framework.templates.template_interface.TemplateInterface.apply_template_for_pipeline_params`.
        """
        self.df: Optional[pd.DataFrame] = None
        """Timeseries data to forecast."""
        self.config: Optional[ForecastConfig] = None
        """Forecast configuration."""
        self.pipeline_params: Optional[Dict] = None
        """Parameters (keyword arguments) to call
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """

    @property
    @abstractmethod
    def allow_model_template_list(self):
        """Whether the template accepts a list for `config.model_template` (bool)"""

    @property
    @abstractmethod
    def allow_model_components_param_list(self):
        """Whether the template accepts a list for `config.model_components_param` (bool)"""

    @abstractmethod
    def apply_template_for_pipeline_params(self, df: pd.DataFrame, config: Optional[ForecastConfig] = None) -> Dict:
        """Converts forecast config to pipeline params.

        Takes input data and optional configuration parameters
        to customize the model. Returns a set of parameters to call
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.

        Parameters
        ----------
        df : `pandas.DataFrame`
            Timeseries data to forecast.
            See :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline` for the format.
            This is the only required parameter.

            ``df`` should only contain time_col, value_col, and regressors.
            Every column in ``df`` besides time_col, value_col is assumed to be a regressor.
            Function does not allow prediction beyond the max available date with any regressor.
        config : :class:`~greykite.framework.templates.autogen.forecast_config.ForecastConfig` or None
            Forecast configuration, including metadata about ``df``,
            forecast horizon, prediction band coverage, evaluation setup,
            model components, and computation. See
            :class:`~greykite.framework.templates.autogen.forecast_config.ForecastConfig`.
            Can be initialized via::

                config = ForecastConfig(
                    model_template=model_template,
                    forecast_horizon=forecast_horizon,
                    coverage=coverage,
                    metadata_param=MetadataParam(...),
                    evaluation_metric_param=EvaluationMetricParam(...),
                    evaluation_period_param=EvaluationPeriodParam(...),
                    model_components_param=ModelComponentsParam(...),
                    computation_param=ComputationParam(...),
                )

        Returns
        -------
        forecast_parameters : `dict` [ `str`, `any` ]
            Parameters (keyword arguments) to call
            :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
        """
        pass



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Utility functions for
`~greykite.framework.pipeline.pipeline`.
"""

import math
from functools import partial

import pandas as pd
from sklearn.base import clone
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import RandomizedSearchCV
from sklearn.pipeline import Pipeline

from greykite.common import constants as cst
from greykite.common.constants import FRACTION_OUTSIDE_TOLERANCE
from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.evaluation import add_finite_filter_to_scorer
from greykite.common.evaluation import add_preaggregation_to_scorer
from greykite.common.evaluation import fraction_outside_tolerance
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import get_integer
from greykite.common.time_properties_forecast import get_default_horizon_from_period
from greykite.framework.constants import CUSTOM_SCORE_FUNC_NAME
from greykite.framework.constants import CV_REPORT_METRICS_ALL
from greykite.framework.constants import FRACTION_OUTSIDE_TOLERANCE_NAME
from greykite.framework.output.univariate_forecast import UnivariateForecast
from greykite.sklearn.estimator.simple_silverkite_estimator import SimpleSilverkiteEstimator
from greykite.sklearn.sklearn_scorer import make_scorer_df
from greykite.sklearn.transform.column_selector import ColumnSelector
from greykite.sklearn.transform.drop_degenerate_transformer import DropDegenerateTransformer
from greykite.sklearn.transform.dtype_column_selector import DtypeColumnSelector
from greykite.sklearn.transform.normalize_transformer import NormalizeTransformer
from greykite.sklearn.transform.null_transformer import NullTransformer
from greykite.sklearn.transform.pandas_feature_union import PandasFeatureUnion
from greykite.sklearn.transform.zscore_outlier_transformer import ZscoreOutlierTransformer


def get_best_index(results, metric="score", greater_is_better=False):
    """Suitable for use as the `refit` parameter to
    `~sklearn.model_selection.RandomizedSearchCV`, after wrapping
    with `functools.partial`.

    Callable that takes ``cv_results_`` from grid search
    and returns the best index.

    Parameters
    ----------
    results : `dict` [`str`, `numpy.array`]
        Results from CV grid search.
        See `~sklearn.model_selection.RandomizedSearchCV`
        ``cv_results_`` attribute for the format.
    metric : `str`, default "score"
        Which metric to use to select the best parameters.
        In single metric evaluation, the metric name should be "score".
        For multi-metric evaluation, the ``scoring`` parameter to
        `~sklearn.model_selection.RandomizedSearchCV` is a dictionary,
        and ``metric`` must be a key of ``scoring``.
    greater_is_better : `bool`, default False
        If True, selects the parameters with highest test values
        for ``metric``. Otherwise, selects those with the lowest
        test values for ``metric``.

    Returns
    -------
    best_index : `int`
        Best index to use for refitting the model.

    Examples
    --------
    >>> from functools import partial
    >>> from sklearn.model_selection import RandomizedSearchCV
    >>> refit = partial(get_best_index, metric="score", greater_is_better=False)
    >>> # RandomizedSearchCV(..., refit=refit)
    """
    if greater_is_better:
        # Note: in case of ties, the index corresponds to the first
        #   optimal value. But the order during CV may not match the order
        #   in fitted grid search ``.cv_results_`` attribute.
        # Note: "rank_test_{metric}" is ranked assuming greater_is_better=True,
        #   so these ranks are the opposite of the true ranks if greater_is_better=False.
        best_index = results[f"mean_test_{metric}"].argmax()
    else:
        best_index = results[f"mean_test_{metric}"].argmin()
    return best_index


def get_default_time_parameters(
        period,
        num_observations,
        forecast_horizon=None,
        test_horizon=None,
        periods_between_train_test=None,
        cv_horizon=None,
        cv_min_train_periods=None,
        cv_expanding_window=False,
        cv_periods_between_splits=None,
        cv_periods_between_train_test=None,
        cv_max_splits=3):
    """Returns default forecast horizon, backtest, and cross-validation parameters,
    given the input frequency, size, and user requested values.

    This function is called from the `~greykite.framework.pipeline.pipeline.forecast_pipeline`
    directly, to provide suitable default to users of forecast_pipeline, and because the default
    should not depend on model configuration (the template).

    Parameters
    ----------
    period: `float`
        Period of each observation (i.e. average time between observations, in seconds).
    num_observations: `int`
        Number of observations in the input data.
    forecast_horizon: `int` or None, default None
        Number of periods to forecast into the future. Must be > 0.
        If None, default is determined from input data frequency.
    test_horizon: `int` or None, default None
        Numbers of periods held back from end of df for test.
        The rest is used for cross validation.
        If None, default is ``forecast_horizon``. Set to 0 to skip backtest.
    periods_between_train_test : `int` or None, default None
        Number of periods gap between train and test in a CV split.
        If None, default is 0.
    cv_horizon: `int` or None, default None
        Number of periods in each CV test set.
        If None, default is ``forecast_horizon``. Set to 0 to skip CV.
    cv_min_train_periods: `int` or None, default None
        Minimum number of periods for training each CV fold.
        If ``cv_expanding_window`` is False, every training period is this size.
        If None, default is 2 * ``cv_horizon``.
    cv_expanding_window: `bool`, default False
        If True, training window for each CV split is fixed to the first available date.
        Otherwise, train start date is sliding, determined by ``cv_min_train_periods``.
    cv_periods_between_splits: `int` or None, default None
        Number of periods to slide the test window between CV splits
        If None, default is ``cv_horizon``.
    cv_periods_between_train_test: `int` or None, default None
        Number of periods gap between train and test in a CV split.
        If None, default is ``periods_between_train_test``.
    cv_max_splits: `int` or None, default 3
        Maximum number of CV splits. Given the above configuration, samples up to max_splits train/test splits,
        preferring splits toward the end of available data. If None, uses all splits.

    Returns
    -------
    time_params : `dict` [`str`, `str`]
        keys are parameter names, values are their default values.
    """
    if forecast_horizon is None:
        forecast_horizon = get_default_horizon_from_period(
            period=period,
            num_observations=num_observations)
    forecast_horizon = get_integer(val=forecast_horizon, name="forecast_horizon", min_value=1)

    test_horizon = get_integer(
        val=test_horizon,
        name="test_horizon",
        min_value=0,
        default_value=forecast_horizon)
    # reduces test_horizon to default 80/20 split if there is not enough data
    if test_horizon >= num_observations:
        test_horizon = math.floor(num_observations * 0.2)

    cv_horizon = get_integer(
        val=cv_horizon,
        name="cv_horizon",
        min_value=0,
        default_value=forecast_horizon)
    # RollingTimeSeriesSplit handles the case of no CV splits, not handled in detail here
    # temporary patch to avoid the case where cv_horizon==num_observations, which throws an error
    # in RollingTimeSeriesSplit
    if cv_horizon >= num_observations:
        cv_horizon = math.floor(num_observations * 0.2)

    periods_between_train_test = get_integer(
        val=periods_between_train_test,
        name="periods_between_train_test",
        min_value=0,
        default_value=0)

    cv_periods_between_train_test = get_integer(
        val=cv_periods_between_train_test,
        name="cv_periods_between_train_test",
        min_value=0,
        default_value=periods_between_train_test)

    return {
        "forecast_horizon": forecast_horizon,
        "test_horizon": test_horizon,
        "periods_between_train_test": periods_between_train_test,
        "cv_horizon": cv_horizon,
        "cv_min_train_periods": cv_min_train_periods,
        "cv_periods_between_train_test": cv_periods_between_train_test
    }


def get_basic_pipeline(
        estimator=SimpleSilverkiteEstimator(),
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        score_func_greater_is_better=False,
        agg_periods=None,
        agg_func=None,
        relative_error_tolerance=None,
        coverage=0.95,
        null_model_params=None,
        regressor_cols=None):
    """Returns a basic pipeline for univariate forecasting.
    Allows for outlier detection, normalization, null imputation,
    degenerate column removal, and forecast model fitting. By default,
    only null imputation is enabled. See source code for the pipeline steps.

    Notes
    -----
    While ``score_func`` is used to define the estimator's score function, the
    the ``scoring`` parameter of `~sklearn.model_selection.RandomizedSearchCV`
    should be provided when using this pipeline in grid search.
    Otherwise, grid search assumes higher values are better for ``score_func``.

    Parameters
    ----------
    estimator : instance of an estimator that implements `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`, default SimpleSilverkiteEstimator()  # noqa: E501
        Estimator to use as the final step in the pipeline.
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    score_func_greater_is_better : `bool`, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.
    agg_periods : `int` or None, default None
        Number of periods to aggregate before evaluation.
        Model is fit at original frequency, and forecast is
        aggregated according to ``agg_periods``
        E.g. fit model on hourly data, and evaluate performance at daily level
        If None, does not apply aggregation
    agg_func : callable or None, default None
        Takes an array and returns a number, e.g. np.max, np.sum
        Used to aggregate data prior to evaluation (applied to actual and predicted)
        Ignored if ``agg_periods`` is None
    relative_error_tolerance : `float` or None, default None
        Threshold to compute the
        `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`
        metric, defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        Required if ``score_func`` is
        `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    coverage : `float` or None, optional, default=0.95
        Intended coverage of the prediction bands (0.0 to 1.0)
        If None, the upper/lower predictions are not returned
        Ignored if `pipeline` is provided. Uses coverage of the ``pipeline`` estimator instead.
    null_model_params : `dict` or None, default None
        Defines baseline model to compute ``R2_null_model_score`` evaluation metric.
        ``R2_null_model_score`` is the improvement in the loss function relative
        to a null model. It can be used to evaluate model quality with respect to
        a simple baseline. For details, see
        `~greykite.common.evaluation.r2_null_model_score`.

        The null model is a `~sklearn.dummy.DummyRegressor`,
        which returns constant predictions.

        Valid keys are "strategy", "constant", "quantile".
        See `~sklearn.dummy.DummyRegressor`. For example::

            null_model_params = {
                "strategy": "mean",
            }
            null_model_params = {
                "strategy": "median",
            }
            null_model_params = {
                "strategy": "quantile",
                "quantile": 0.8,
            }
            null_model_params = {
                "strategy": "constant",
                "constant": 2.0,
            }

        If None, ``R2_null_model_score`` is not calculated.

        Note: CV model selection always optimizes ``score_func`, not
        the ``R2_null_model_score``.

    regressor_cols : `list` [`str`] or None, default None
        A list of regressor columns used in the training and prediction DataFrames.
        It should contain only the regressors that are being used in the grid search.
        If None, no regressor columns are used.
        Regressor columns that are unavailable in ``df`` are dropped.

    Returns
    -------
    pipeline : `sklearn.pipeline.Pipeline`
        sklearn Pipeline for univariate forecasting.
    """
    score_func, _, _ = get_score_func_with_aggregation(
        score_func=score_func,
        greater_is_better=score_func_greater_is_better,
        agg_periods=agg_periods,
        agg_func=agg_func,
        relative_error_tolerance=relative_error_tolerance)

    if regressor_cols is None:
        regressor_cols = []

    # A new unfitted estimator with the same parameters
    estimator_clone = clone(estimator)
    # Sets parameters common to all `BaseForecastEstimator`
    estimator_clone.set_params(
        score_func=score_func,
        coverage=coverage,
        null_model_params=null_model_params)

    # Note:
    #   Unlike typical ML, "y" (target values) is part of "X" (training data).
    #   Some forecasting models require that all historical values are available
    #   (no gaps in the timeseries). By including "y" values as part of "X", they
    #   can be transformed prior to fitting the estimator. This allows outlier removal and
    #   null imputation that respects train/test boundaries, to avoid leaking future
    #   information into the past. Evaluation is always done against original "y".
    #   Parameters for this pipeline are set via `hyperparameter_grid`.
    pipeline = Pipeline([
        ("input", PandasFeatureUnion([
            ("date", Pipeline([
                ("select_date", ColumnSelector([TIME_COL]))  # leaves time column unmodified
            ])),
            ("response", Pipeline([  # applies outlier and null transformation to value column
                ("select_val", ColumnSelector([VALUE_COL])),
                ("outlier", ZscoreOutlierTransformer(z_cutoff=None)),
                ("null", NullTransformer(impute_algorithm="interpolate"))
            ])),
            ("regressors_numeric", Pipeline([
                ("select_reg", ColumnSelector(regressor_cols)),
                ("select_reg_numeric", DtypeColumnSelector(include="number")),
                ("outlier", ZscoreOutlierTransformer(z_cutoff=None)),
                ("normalize", NormalizeTransformer(normalize_algorithm=None)),  # no normalization by default
                ("null", NullTransformer(impute_algorithm="interpolate"))
            ])),
            ("regressors_other", Pipeline([
                ("select_reg", ColumnSelector(regressor_cols)),
                ("select_reg_non_numeric", DtypeColumnSelector(exclude="number"))
            ]))
        ])),
        ("degenerate", DropDegenerateTransformer()),  # default `drop_degenerate=False`
        # Sets BaseForecastEstimator parameters (`score_func`, etc.).
        # Other parameters of the estimator are set by `hyperparameter_grid` later.
        ("estimator", estimator_clone)
    ])
    return pipeline


def get_score_func_with_aggregation(
        score_func,
        greater_is_better=None,
        agg_periods=None,
        agg_func=None,
        relative_error_tolerance=None):
    """Returns a score function that pre-aggregates inputs according to ``agg_func``,
    and filters out invalid true values before evaluation. This allows fitting
    the model at a granular level, yet evaluating at a coarser level.

    Also returns the proper direction and short name for the score function.

    Parameters
    ----------
    score_func : `str` or callable
        If callable, a function that maps two arrays to a number:
        ``(true, predicted) -> score``.
    greater_is_better : `bool`, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.
    agg_periods : `int` or None, default None
        Number of periods to aggregate before evaluation.
        Model is fit at original frequency, and forecast is
        aggregated according to ``agg_periods``
        E.g. fit model on hourly data, and evaluate performance at daily level
        If None, does not apply aggregation
    agg_func : callable or None, default None
        Takes an array and returns a number, e.g. np.max, np.sum
        Used to aggregate data prior to evaluation (applied to actual and predicted)
        Ignored if ``agg_periods`` is None
    relative_error_tolerance : `float` or None, default None
        Threshold to compute the
        `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`
        metric, defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        Required if ``score_func`` is
        `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

    Returns
    -------
    score_func : callable
        scorer with pre-aggregation function and filter,
    greater_is_better : `bool`
        Whether ``greater_is_better`` for the scorer.
        Uses the provided ``greater_is_better`` if the
        provided ``score_func`` is a callable.
        Otherwise, looks up the direction.
    short_name : `str`
        Canonical short name for the ``score_func``.
    """
    if isinstance(score_func, str):
        # If a string is provided, looks up the callable score_func and greater_is_better.
        # Otherwise, uses the args directly
        if score_func == FRACTION_OUTSIDE_TOLERANCE:
            if relative_error_tolerance is None:
                raise ValueError("Must specify `relative_error_tolerance` to request "
                                 "FRACTION_OUTSIDE_TOLERANCE as a metric.")
            score_func = partial(
                fraction_outside_tolerance,
                rtol=relative_error_tolerance)
            greater_is_better = False
            short_name = FRACTION_OUTSIDE_TOLERANCE_NAME
        else:
            try:
                enum = EvaluationMetricEnum[score_func]
                score_func = enum.get_metric_func()
                greater_is_better = enum.get_metric_greater_is_better()
                short_name = enum.get_metric_name()
            except KeyError:
                valid_names = ", ".join(EvaluationMetricEnum.__dict__["_member_names_"])
                raise NotImplementedError(f"Evaluation metric {score_func} is not available. Must be one of: {valid_names}")
    elif callable(score_func):
        # Uses the score_func and greater_is_better passed to the function
        short_name = CUSTOM_SCORE_FUNC_NAME
    else:
        raise ValueError("`score_func` must be an `EvaluationMetricEnum` member name, "
                         "FRACTION_OUTSIDE_TOLERANCE, or callable.")

    if agg_periods is not None:
        score_func = add_preaggregation_to_scorer(score_func, agg_periods, agg_func)

    # Filters out elements that can't be compared in ``y_true``
    score_func = add_finite_filter_to_scorer(score_func)

    return score_func, greater_is_better, short_name


def get_scoring_and_refit(
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        score_func_greater_is_better=False,
        cv_report_metrics=None,
        agg_periods=None,
        agg_func=None,
        relative_error_tolerance=None):
    """Provides ``scoring`` and ``refit`` parameters for
    `~sklearn.model_selection.RandomizedSearchCV`.

    Together, ``scoring`` and ``refit`` specify how what metrics to evaluate and how
    to evaluate the predictions on the test set to identify the optimal model.

    Notes
    -----
    Sets ``greater_is_better=True`` in `scoring` for all metrics to report them with their
    original sign, and properly accounts for this in ``refit`` to extract the best index.

    Pass both `scoring` and `refit` to `~sklearn.model_selection.RandomizedSearchCV`

    Parameters
    ----------
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    score_func_greater_is_better : `bool`, default False
        True if ``score_func`` is a score function, meaning higher is better,
        and False if it is a loss function, meaning lower is better.
        Must be provided if ``score_func`` is a callable (custom function).
        Ignored if ``score_func`` is a string, because the direction is known.
    cv_report_metrics : `~greykite.common.constants.CV_REPORT_METRICS_ALL`, or `list` [`str`], or None, default None  # noqa: E501
        Additional metrics to compute during CV, besides the one specified by ``score_func``.

            - If the string constant `greykite.common.constants.CV_REPORT_METRICS_ALL`,
              computes all metrics in ``EvaluationMetricEnum``. Also computes
              ``FRACTION_OUTSIDE_TOLERANCE`` if ``relative_error_tolerance`` is not None.
              The results are reported by the short name (``.get_metric_name()``) for ``EvaluationMetricEnum``
              members and ``FRACTION_OUTSIDE_TOLERANCE_NAME`` for ``FRACTION_OUTSIDE_TOLERANCE``.
            - If a list of strings, each of the listed metrics is computed. Valid strings are
              `greykite.common.evaluation.EvaluationMetricEnum` member names
              and `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.

              For example::

                ["MeanSquaredError", "MeanAbsoluteError", "MeanAbsolutePercentError", "MedianAbsolutePercentError", "FractionOutsideTolerance2"]

            - If None, no additional metrics are computed.
    agg_periods : `int` or None, default None
        Number of periods to aggregate before evaluation.
        Model is fit at original frequency, and forecast is
        aggregated according to ``agg_periods``
        E.g. fit model on hourly data, and evaluate performance at daily level
        If None, does not apply aggregation
    agg_func : callable or None, default None
        Takes an array and returns a number, e.g. np.max, np.sum
        Used to aggregate data prior to evaluation (applied to actual and predicted)
        Ignored if ``agg_periods`` is None
    relative_error_tolerance : `float` or None, default None
        Threshold to compute the
        `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`
        metric, defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        If `None`, the metric is not computed.

    Returns
    -------
    scoring : `dict`
        A dictionary of metrics to evaluate for each CV split.
        The key is the metric name, the value is an instance
        of `~greykite.common.evaluation_PredictScorerDF` generated
        by :func:`~greykite.common.evaluation.make_scorer_df`.

        The value has a score method that takes actual and predicted values
        and returns a single number.

        There is one item in the dictionary for ``score_func``
        and an additional item for each additional element in
        ``cv_report_metrics``.

            - The key for ``score_func`` if it is a callable is
              `~greykite.common.constants.CUSTOM_SCORE_FUNC_NAME`.
            - The key for ``EvaluationMetricEnum`` member name is the short name
              from ``.get_metric_name()``.
            - The key for `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`
              is `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE_NAME`.

        See `~sklearn.model_selection.RandomizedSearchCV`.

    refit : callable
        Callable that takes ``cv_results_`` from grid search
        and returns the best index.

        See `~sklearn.model_selection.RandomizedSearchCV`.
    """
    if cv_report_metrics is None:
        cv_report_metrics = []
    elif cv_report_metrics == CV_REPORT_METRICS_ALL:
        cv_report_metrics = EvaluationMetricEnum.__dict__["_member_names_"].copy()
        # Computes `FRACTION_OUTSIDE_TOLERANCE` if `relative_error_tolerance` is specified
        if relative_error_tolerance is not None:
            cv_report_metrics.append(FRACTION_OUTSIDE_TOLERANCE)
    else:
        cv_report_metrics = cv_report_metrics.copy()

    # Defines scoring metrics to evaluate on each CV split.
    # The results in .cv_results_ attribute are reported as
    # ``f"mean_test_{name}"``, ``f"mean_train_{name}"``, etc.,
    # where `name` is a key in the `scoring` dictionary.
    scoring = {}
    # Adds all recognized metrics in `cv_report_metrics` to `scoring`
    for metric_name in cv_report_metrics:
        # Adds aggregation to the metric function and gets the short name
        func, _, short_name = get_score_func_with_aggregation(
            score_func=metric_name,
            agg_periods=agg_periods,
            agg_func=agg_func,
            relative_error_tolerance=relative_error_tolerance)
        # Strings/functions typically accepted by
        #   `sklearn.model_selection.RandomizedSearchCV`
        #   are not compatible with the predictions returned by
        #   `~greykite.sklearn.estimator.base_forecast_estimator.BaseForecastEstimator`
        #   because predictions are returned as a dataframe.
        #   Thus, instead of ``"neg_median_absolute_error"``,
        #   use ``make_scorer_df(median_absolute_error, greater_is_better=False)``
        #   See `~greykite.common.evaluation.make_scorer_df` for details.
        # Uses `greater_is_better=True` to avoid flipping the metric sign in CV results.
        scoring[short_name] = make_scorer_df(
            score_func=func,
            greater_is_better=True)

    # Adds `score_func` to `scoring`
    func, greater_is_better, short_name = get_score_func_with_aggregation(
        score_func=score_func,  # string or callable
        greater_is_better=score_func_greater_is_better,
        agg_periods=agg_periods,
        agg_func=agg_func,
        relative_error_tolerance=relative_error_tolerance)
    scoring[short_name] = make_scorer_df(
        score_func=func,
        # Uses `greater_is_better=True` to avoid flipping the metric sign in CV results.
        # The `refit` parameter defined below should be passed to grid search to ensure proper
        # extraction of the optimal result, accounting for `greater_is_better` of the metric.
        greater_is_better=True)
    # Defines `refit` function with the true `greater_is_better` to pick the best result from CV.
    refit = partial(get_best_index, metric=short_name, greater_is_better=greater_is_better)
    return scoring, refit


def get_hyperparameter_searcher(
        hyperparameter_grid,
        model,
        cv=None,
        hyperparameter_budget=None,
        n_jobs=1,
        verbose=1,
        **kwargs) -> RandomizedSearchCV:
    """Returns RandomizedSearchCV object for hyperparameter tuning via cross validation

    `sklearn.model_selection.RandomizedSearchCV` runs a full grid search if
    ``hyperparameter_budget`` is sufficient to exhaust the full
    ``hyperparameter_grid``, otherwise it samples uniformly at random from the space.

    Parameters
    ----------
    hyperparameter_grid : `dict` or `list` [`dict`]
        Dictionary with parameters names (string) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        Lists of parameters are sampled uniformly.

        May also be a list of such dictionaries to avoid undesired combinations of parameters.
        Passed as ``param_distributions`` to `sklearn.model_selection.RandomizedSearchCV`,
        see docs for more info.
    model: estimator object
        A object of that type is instantiated for each grid point. This is assumed to implement
        the scikit-learn estimator interface.
    cv: `int`, cross-validation generator, iterable, or None, default None
        Determines the cross-validation splitting strategy.
        See `sklearn.model_selection.RandomizedSearchCV`.
    hyperparameter_budget: `int` or None, default None
        max number of hyperparameter sets to try within the hyperparameter_grid search space
        If None, uses defaults:

            * exhaustive grid search if all values are constant
            * 10 if any value is a distribution to sample from

    n_jobs : `int` or None, default 1
        Number of jobs to run in parallel
        (the maximum number of concurrently running workers).
        ``-1`` uses all CPUs. ``-2`` uses all CPUs but one.
        ``None`` is treated as 1 unless in a `joblib.Parallel` backend context
        that specifies otherwise.
    verbose : `int`, default 1
        Verbosity level during CV.

        * if > 0, prints number of fits
        * if > 1, prints fit parameters, total score + fit time
        * if > 2, prints train/test scores
    kwargs : additional parameters
        Keyword arguments to pass to `~greykite.framework.pipeline.utils.get_scoring_and_refit`.
        Accepts the following parameters:

            - ``"score_func"``
            - ``"score_func_greater_is_better"``
            - ``"cv_report_metrics"``
            - ``"agg_periods"``
            - ``"agg_func"``
            - ``"relative_error_tolerance"``

    Returns
    -------
    grid_search : `sklearn.model_selection.RandomizedSearchCV`
        Object that can run randomized search on hyper parameters.
    """
    if hyperparameter_budget is None:
        # sets reasonable defaults when hyperparameter_budget is not provided
        try:
            # exhaustive search if explicit values are provided
            hyperparameter_budget = len(ParameterGrid(hyperparameter_grid))
            log_message(f"Setting hyperparameter_budget to {hyperparameter_budget} for full grid search.",
                        LoggingLevelEnum.DEBUG)
        except TypeError:  # parameter value is not iterable
            # sets budget to 10 if distribution for randomized search is provided
            hyperparameter_budget = 10
            log_message(f"Setting hyperparameter_budget to {hyperparameter_budget} to sample from"
                        f" provided distributions (and lists).", LoggingLevelEnum.WARNING)

    scoring, refit = get_scoring_and_refit(**kwargs)

    # note: RandomizedSearchCV operates like GridSearchCV when hyperparameter_grid contains no distributions
    grid_search = RandomizedSearchCV(
        estimator=model,
        param_distributions=hyperparameter_grid,  # a fixed list or distribution to sample from
        n_iter=hyperparameter_budget,             # samples uniformly, up to hyperparameter_budget
        scoring=scoring,                          # model evaluation criteria (note: if None, uses the score function of the estimator)
        n_jobs=n_jobs,                            # parallelism
        refit=refit,                              # selects the best model
        cv=cv,
        verbose=verbose,
        pre_dispatch="2*n_jobs",                  # controls memory consumption
        return_train_score=True                   # NB: could be False for speedup
    )
    return grid_search


def get_forecast(
        df,
        trained_model: Pipeline,
        train_end_date=None,
        test_start_date=None,
        forecast_horizon=None,
        xlabel=cst.TIME_COL,
        ylabel=cst.VALUE_COL,
        relative_error_tolerance=None) -> UnivariateForecast:
    """Runs model predictions on ``df`` and creates a
    `~greykite.framework.output.univariate_forecast.UnivariateForecast` object.

    Parameters
    ----------
    df: `pandas.DataFrame`
        Has columns cst.TIME_COL, cst.VALUE_COL, to forecast.
    trained_model: `sklearn.pipeline`
        A fitted Pipeline with ``estimator`` step and predict function.
    train_end_date: `datetime.datetime`, default `None`
        Train end date. Passed to
        `~greykite.framework.output.univariate_forecast.UnivariateForecast`.
    test_start_date: `datetime.datetime`, default `None`
        Test start date. Passed to
        `~greykite.framework.output.univariate_forecast.UnivariateForecast`.
    forecast_horizon : `int` or None, default None
        Number of periods forecasted into the future. Must be > 0. Passed to
        `~greykite.framework.output.univariate_forecast.UnivariateForecast`.
    xlabel: `str`
        Time column to use in representing forecast (e.g. x-axis in plots).
    ylabel: `str`
        Time column to use in representing forecast (e.g. y-axis in plots).
    relative_error_tolerance : `float` or None, default None
        Threshold to compute the ``Outside Tolerance`` metric,
        defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        If `None`, the metric is not computed.

    Returns
    -------
    univariate_forecast : `~greykite.framework.output.univariate_forecast.UnivariateForecast`
        Forecasts represented as a ``UnivariateForecast`` object.
    """
    predicted_df = trained_model.predict(df)
    # This is more robust than using trained_model.named_steps["estimator"] e.g.
    # if the user calls forecast_pipeline with a custom pipeline, where the last
    # step isn't named "estimator".
    trained_estimator = trained_model.steps[-1][-1]
    coverage = trained_estimator.coverage

    # combines actual with predictions
    union_df = pd.DataFrame({
        xlabel: df[cst.TIME_COL].values,
        # .values here, since df and predicted_df have different indexes
        cst.ACTUAL_COL: df[cst.VALUE_COL].values,
        # evaluation and plots are done on the values *before* any transformations
        cst.PREDICTED_COL: predicted_df[cst.PREDICTED_COL].values
    })

    predicted_lower_col = None
    predicted_upper_col = None
    null_model_predicted_col = None

    # adds lower bound if available
    if cst.PREDICTED_LOWER_COL in predicted_df.columns:
        predicted_lower_col = cst.PREDICTED_LOWER_COL
        union_df[cst.PREDICTED_LOWER_COL] = predicted_df[cst.PREDICTED_LOWER_COL].values
        if coverage is None:
            raise Exception("coverage must be provided")

    # adds upper bound if available
    if cst.PREDICTED_UPPER_COL in predicted_df.columns:
        predicted_upper_col = cst.PREDICTED_UPPER_COL
        union_df[cst.PREDICTED_UPPER_COL] = predicted_df[cst.PREDICTED_UPPER_COL].values
        if coverage is None:
            raise Exception("coverage must be provided")

    # adds null prediction if available
    if trained_estimator.null_model is not None:
        null_model_predicted_col = cst.NULL_PREDICTED_COL
        null_predicted_df = trained_estimator.null_model.predict(df)
        union_df[cst.NULL_PREDICTED_COL] = null_predicted_df[cst.PREDICTED_COL].values

    return UnivariateForecast(
        union_df,
        time_col=xlabel,
        actual_col=cst.ACTUAL_COL,
        predicted_col=cst.PREDICTED_COL,
        predicted_lower_col=predicted_lower_col,
        predicted_upper_col=predicted_upper_col,
        null_model_predicted_col=null_model_predicted_col,
        ylabel=ylabel,
        train_end_date=train_end_date,
        test_start_date=test_start_date,
        forecast_horizon=forecast_horizon,
        coverage=coverage,
        r2_loss_function=trained_estimator.score_func,  # this score_func includes preaggregation if requested
        estimator=trained_estimator,
        relative_error_tolerance=relative_error_tolerance
    )



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Constants used by code in `~greykite.common` or in multiple places:
`~greykite.algo`, `~greykite.sklearn`,
and/or `~greykite.framework`.
"""

# The time series data is represented in pandas dataframes
# The default column names for the series are given below
TIME_COL = "ts"
"""The default name for the column with the timestamps of the time series"""
VALUE_COL = "y"
"""The default name for the column with the values of the time series"""
ACTUAL_COL = "actual"
"""The column name representing actual (observed) values"""
PREDICTED_COL = "forecast"
"""The column name representing the predicted values"""
PREDICTED_LOWER_COL = "forecast_lower"
"""The column name representing upper bounds of prediction interval"""
PREDICTED_UPPER_COL = "forecast_upper"
"""The column name representing lower bounds of prediction interval"""
NULL_PREDICTED_COL = "forecast_null"
"""The column name representing predicted values from null model"""
ERR_STD_COL = "err_std"
"""The column name representing the error standard deviation from models"""

# Evaluation metrics corresponding to `~greykite.common.evaluation`
R2_null_model_score = "R2_null_model_score"
"""Evaluation metric. Improvement in the specified loss function compared to the predictions of a null model."""
FRACTION_OUTSIDE_TOLERANCE = "Outside Tolerance (fraction)"
"""Evaluation metric. The fraction of predictions outside the specified tolerance level"""
PREDICTION_BAND_WIDTH = "Prediction Band Width (%)"
"""Evaluation metric. Relative size of prediction bands vs actual, as a percent"""
PREDICTION_BAND_COVERAGE = "Prediction Band Coverage (fraction)"
"""Evaluation metric. Fraction of observations within the bands"""
LOWER_BAND_COVERAGE = "Coverage: Lower Band"
"""Evaluation metric. Fraction of observations within the lower band"""
UPPER_BAND_COVERAGE = "Coverage: Upper Band"
"""Evaluation metric. Fraction of observations within the upper band"""
COVERAGE_VS_INTENDED_DIFF = "Coverage Diff: Actual_Coverage - Intended_Coverage"
"""Evaluation metric. Difference between actual and intended coverage"""

# Column names used by `~greykite.common.features.timeseries_features`
EVENT_DF_DATE_COL = "date"
"""Name of date column for the DataFrames passed to silverkite `custom_daily_event_df_dict`"""
EVENT_DF_LABEL_COL = "event_name"
"""Name of event column for the DataFrames passed to silverkite `custom_daily_event_df_dict`"""
EVENT_PREFIX = "events"
"""Prefix for naming event features."""
EVENT_DEFAULT = ""
"""Label used for days without an event."""
EVENT_INDICATOR = "event"
"""Binary indicatory for an event"""
CHANGEPOINT_COL_PREFIX = "changepoint"
"""Prefix for naming changepoint features."""
CHANGEPOINT_COL_PREFIX_SHORT = "cp"
"""Short prefix for naming changepoint features."""

# Column names used by
# `~greykite.common.features.adjust_anomalous_data.adjust_anomalous_data`
START_DATE_COL = "start_date"
"""Start timestamp column name"""
END_DATE_COL = "end_date"
"""Standard end timestamp column"""
ADJUSTMENT_DELTA_COL = "adjustment_delta"
"""Adjustment column"""
METRIC_COL = "metric"
"""Column to denote metric of interest"""
DIMENSION_COL = "dimension"
"""Dimension column"""

# Constants related to
# `~greykite.common.features.timeseries_features.build_time_features_df`.
GROWTH_COL_ALIAS = {
    "linear": "ct1",
    "quadratic": "ct2",
    "cubic": "ct3",
    "sqrt": "ct_sqrt",
    "cuberoot": "ct_root3"
}
"""Human-readable names for the growth columns generated by
`~greykite.common.features.timeseries_features.build_time_features_df`
"""
TIME_FEATURES = [
    "datetime", "date",
    "year", "year_length",
    "quarter", "quarter_start", "quarter_length",
    "month", "month_length",
    "woy", "doy", "doq", "dom", "dow",
    "str_dow", "str_doy",
    "hour", "minute", "second",
    "year_month",
    "year_woy", "month_dom", "year_woy_dow", "woy_dow", "dow_hr", "dow_hr_min",
    "tod", "tow", "tom", "toq", "toy",
    "conti_year", "is_weekend", "dow_grouped",
    "ct1", "ct2", "ct3", "ct_sqrt", "ct_root3"]
"""Time features generated by
`~greykite.common.features.timeseries_features.build_time_features_df`"""

# Column names used by
# `~greykite.common.features.timeseries_lags`
LAG_INFIX = "_lag"
"""Infix for lagged feature names"""
AGG_LAG_INFIX = "avglag"
"""Infix for aggregated lag feature names"""

# Patterns for categorizing timeseries features
TREND_REGEX = f"{CHANGEPOINT_COL_PREFIX}\\d|ct\\d|ct_|{CHANGEPOINT_COL_PREFIX_SHORT}\\d"
"""Growth terms, including changepoints."""
SEASONALITY_REGEX = "sin\\d|cos\\d"
"""Seasonality terms modeled by fourier series."""
EVENT_REGEX = f"{EVENT_PREFIX}_"
"""Event terms."""
LAG_REGEX = f"{LAG_INFIX}\\d|_{AGG_LAG_INFIX}_\\d"
"""Lag terms."""

LOGGER_NAME = "Greykite"
"""Name used by the logger."""



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Kaixu Yang
"""Utilities for silverkite feature column names."""

import re

from greykite.common import constants as cst


INTERCEPT = "Intercept"
"""The name of the intercept term, to show in model summary."""


def simplify_event(name):
    """Simplifies an event predictor name.

    Eliminates the levels and unnecessary characters in an event predictor's name.
    If the original name is
    "C(Q('events_Chinese New Year_minus_1'), levels=['', 'event'])[T.event]",
    it will be simplified to "events_Chinese New Year-1".
    If ``cst.EVENT_PREFIX`` is not in the name, the original name will be returned.

    Parameters
    ----------
    name : `str`
        The predictor name.

    Returns
    -------
    simplified_name : `str`
        The simplified name.
    """
    if cst.EVENT_PREFIX in name:
        result = re.search(fr"\('({cst.EVENT_PREFIX}.*)'\),", name)
        if result is not None:
            name = result.group(1)
            name = name.replace("_minus_", "-")
            name = name.replace("_plus_", "+")
    return name


def simplify_time_features(name):
    """Simplifies a time feature predictor name.

    Eliminates the levels and unnecessary characters in a time feature predictor's name.
    If the original name is like
    "C(Q('str_dow'), levels=['1-Mon', '2-Tue', '3-Wed', '4-Thu', '5-Fri', '6-Sat', '7-Sun'])[T.2-Tue]",
    it will be simplified to "str_dow_2-Tue".
    If the original name is like "str_dow[T.7-Sun]", it will be simplified to "str_dow_7-Sun".
    If the original name is like "toy", it will be kept as it is.
    If any ``cst.TIME_FEATURES`` is not in the name, the original name will be returned.

    Parameters
    ----------
    name : `str`
        The predictor name.

    Returns
    -------
    simplified_name : `str`
        The simplified name.
    """
    if any([x in name for x in cst.TIME_FEATURES]):
        result = re.search(r"(.*)\[T.(.*)\]", name)
        if result is not None:
            name = result.group(1)
            level = "_" + result.group(2)
            if "levels" in name:
                name_result = re.search(r"\('(.*)'\),", name)
                if name_result is not None:
                    name = name_result.group(1)
        else:
            level = ""
        if name == "is_weekend":  # is_weekend only has two levels
            level = ""
        name = name + level
    return name


def simplify_changepoints(name):
    """Simplifies a changepoint predictor name.

    Changes "changepoint" to "cp" to shorten the changepoint features' names.
    If ``cst.CHANGEPOINT_COL_PREFIX`` is not in the name, the original name will be returned.

    Parameters
    ----------
    name : `str`
        The predictor name.

    Returns
    -------
    simplified_name : `str`
        The simplified name.
    """
    name = name.replace(cst.CHANGEPOINT_COL_PREFIX, cst.CHANGEPOINT_COL_PREFIX_SHORT)
    return name


def simplify_name(name):
    """Simplifies the predictor names to make them shorter.


    Parameters
    ----------
    name : `str`
        The predictor name. Should be a simple term with no interaction.

    Returns
    -------
    simplified_name : `str`
        The simplified name.
    """
    name = simplify_event(name)
    name = simplify_time_features(name)
    name = simplify_changepoints(name)
    return name


def simplify_pred_cols(pred_cols):
    """Simplifies predictor names in a list.

    Parameters
    ----------
    pred_cols : `list` [ `str` ]
        A list of predictor names to be simplified.
        Names in ``pred_cols`` could contain interactions.

    Returns
    -------
    new_pred_cols : `list` [ `str` ]
        A simplified list of predictor names by applying element-wisely
        `~greykite.algo.common.col_name_utils.simplify_name`
    """
    pred_cols_extract = [[col] if ":" not in col else col.split(":") for col in pred_cols]
    pred_cols_extract = [[simplify_name(name) for name in col] for col in pred_cols_extract]
    new_pred_cols = [col[0] if len(col) == 1 else ":".join(col) for col in pred_cols_extract]
    return new_pred_cols


def add_category_cols(coef_summary, pred_category):
    """Adds indicators columns to coefficient summary df for categaries.

    Parameters
    ----------
    coef_summary : `pandas.DataFrame`
        The coefficient summary df.
        This is typically generated by
        `~greykite.algo.common.model_summary_utils.add_model_coef_df`
        or
        `~greykite.algo.common.model_summary_utils.add_model_coef_df`.
    pred_category : `dict`
        The predictor category dictionary by
        `~greykite.algo.common.col_name_utils.create_pred_category`.

    Returns
    -------
    coef_summary_with_new_columns : `pandas.DataFrame
        New df with the following columns added:

            "is_intercept" : 0 or 1
                Intercept or not.
            "is_time_feature" : 0 or 1
                Time features or not.
                Time features belong to `~greykite.common.constants.TIME_FEATURES`.
            "is_event" : 0 or 1
                Event features or not.
                Event features have `~greykite.common.constants.EVENT_PREFIX`.
            "is_trend" : 0 or 1
                Trend features or not.
                Trend features have `~greykite.common.constants.CHANGEPOINT_COL_PREFIX` or "cp\\d".
            "is_seasonality" : 0 or 1
                Seasonality feature or not.
                Seasonality features have `~greykite.common.constants.SEASONALITY_REGEX`.
            "is_lag" : 0 or 1
                Lagged features or not.
                Lagged features have "lag_".
            "is_regressor" : 0 or 1
                Extra features provided by users.
                They are provided through ``extra_pred_cols`` in the fit function.
            "is_interaction" : 0 or 1
                Interaction feature or not.
                Interaction features have ":".
    """
    coef_summary = coef_summary.copy()
    pred_cols = coef_summary["Pred_col"].tolist()
    coef_summary["is_intercept"] = [1 if col in pred_category["intercept"]
                                    else 0 for col in pred_cols]
    coef_summary["is_time_feature"] = [1 if col in pred_category["time_features"]
                                       else 0 for col in pred_cols]
    coef_summary["is_event"] = [1 if col in pred_category["event_features"]
                                else 0 for col in pred_cols]
    coef_summary["is_trend"] = [1 if col in pred_category["trend_features"]
                                else 0 for col in pred_cols]
    coef_summary["is_seasonality"] = [1 if col in pred_category["seasonality_features"]
                                      else 0 for col in pred_cols]
    coef_summary["is_lag"] = [1 if col in pred_category["lag_features"]
                              else 0 for col in pred_cols]
    coef_summary["is_regressor"] = [1 if col in pred_category["regressor_features"]
                                    else 0 for col in pred_cols]
    coef_summary["is_interaction"] = [1 if col in pred_category["interaction_features"]
                                      else 0 for col in pred_cols]
    return coef_summary


def create_pred_category(pred_cols, extra_pred_cols):
    """Creates a dictionary of predictor categories.

    The keys are categories, and the values are the corresponding
    predictor names. For detail, see
    `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator.pred_category`

    Parameters
    ----------
    pred_cols : `list` [ `str` ]
        A full list of predictor names used in the model, including extra predictor names.
    extra_pred_cols : `list` [ `str` ]
        A list of extra predictors what are manually provided for the estimator class.
        In ``SilverkiteEstimator``, this is the ``extra_pred_cols``.
        In ``SimpleSilverkiteEstimator``, this is the combination of ``regressor_cols``
        and ``extra_pred_cols``.

    Returns
    -------
    pred_category : `dict`
        A dictionary of categories and predictors. For details, see
        `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator.pred_category`
    """

    if extra_pred_cols is None:
        extra_pred_cols = []
    extra_pred_cols = list(set(extra_pred_cols))  # might have duplicates
    pred_category = {
        "intercept": [col for col in pred_cols if re.search(INTERCEPT, col)],
        # Time feature names could be included in seasonality features.
        # We do not want to include pure seasonality features.
        # If a term does not include interaction, it need to include
        # time feature name but not seasonality regex.
        # This keeps "ct1" and excludes "cos1_ct1_yearly"
        # If a term includes interaction, then at least one of the
        # two sub-terms need to satisfy the condition above.
        # This keeps "ct1:is_weekend" and "ct1:cos1_ct1_yearly"
        # and excludes "is_weekend:cos1_ct1_yearly".
        "time_features": [col for col in pred_cols
                          if (re.search(":", col) is None
                              and
                              re.search("|".join(cst.TIME_FEATURES), col)
                              and
                              re.search(cst.SEASONALITY_REGEX, col) is None)
                          or
                          (re.search(":", col)
                           and
                           any([re.search("|".join(cst.TIME_FEATURES), subcol)
                                and
                                re.search(cst.SEASONALITY_REGEX, subcol) is None
                                for subcol in col.split(":")]))
                          ],
        "event_features": [col for col in pred_cols if cst.EVENT_PREFIX in col],
        # the same logic as time features for trend features
        "trend_features": [col for col in pred_cols
                           if (re.search(":", col) is None
                               and
                               re.search(cst.TREND_REGEX, col)
                               and
                               re.search(cst.SEASONALITY_REGEX, col) is None)
                           or
                           (re.search(":", col)
                            and
                            any([re.search(cst.TREND_REGEX, subcol)
                                 and
                                 re.search(cst.SEASONALITY_REGEX, subcol) is None
                                 for subcol in col.split(":")]))
                           ],
        "seasonality_features": [col for col in pred_cols if re.search(cst.SEASONALITY_REGEX, col)],
        "lag_features": [col for col in pred_cols if re.search(cst.LAG_REGEX, col)],
        # only the supplied extra_pred_col that are also in pred_cols
        "regressor_features": [col for col in pred_cols if col in extra_pred_cols],
        "interaction_features": [col for col in pred_cols if re.search(":", col)]
    }
    # removes the regressor_features that are in the other categories
    time_series_features = []
    for category in pred_category:
        if category not in ["regressor_features", "interaction_features"]:
            time_series_features += pred_category[category]
    pred_category["regressor_features"] = [x for x in pred_category["regressor_features"]
                                           if (re.search(":", x) is None
                                               and
                                               x not in time_series_features)
                                           or
                                           (re.search(":", x)
                                            and
                                            any([subx not in time_series_features
                                                 for subx in x.split(":")]))]
    return pred_category


def filter_coef_summary(
        coef_summary,
        pred_category,
        is_intercept=None,
        is_time_feature=None,
        is_event=None,
        is_trend=None,
        is_seasonality=None,
        is_lag=None,
        is_regressor=None,
        is_interaction=None):
    """Gets the coefficient summary df after applying the given filters.

    Set any of the parameters to `bool` to enable filtering.

        - Any argument set to True will be aggregated with logical operator "or", i.e.
          a category will be displayed when set to True.
        - Any argument set to False will be aggregated with logical operator "and", i.e.
          a category will not be displayed when set to False (even it has interaction with
          a category that is set to True).
        - Any argument set to None will be ignored unless all arguments are None.
        - ``is_interaction`` is used to exclude interaction terms by setting it
          to False. It is not used when the value is True and other argument is not None.
          The design here is to use ``is_interaction`` as a second pass filter.

    Parameters
    ----------
    coef_summary : `pandas.DataFrame`
        The coefficient summary df.
    pred_category : `dict`
        The predictor category dictionary by
        `~greykite.algo.common.col_name_utils.create_pred_category`.
    is_intercept : `bool` or `None`, default `None`
        Intercept or not.
    is_time_feature : `bool` or `None`, default `None`
        Time features or not.
        Time features belong to `~greykite.common.constants.TIME_FEATURES`.
    is_event : `bool` or `None`, default `None`
        Event features or not.
        Event features have `~greykite.common.constants.EVENT_PREFIX`.
    is_trend : `bool` or `None`, default `None`
        Trend features or not.
        Trend features have `~greykite.common.constants.CHANGEPOINT_COL_PREFIX`.
    is_seasonality : `bool` or `None`, default `None`
        Seasonality feature or not.
        Seasonality features have `~greykite.common.constants.SEASONALITY_REGEX`.
    is_lag : `bool` or `None`, default `None`
        Lagged features or not.
        Lagged features have "lag_".
    is_regressor : `bool` or `None`, default `None`
        User supplied regressor features or not.
        They are provided with the `extra_pred_cols`.
    is_interaction : `bool` or `None`, default `None`
        Interaction feature or not.
        Interaction features have ":".

    Returns
    -------
    filtered_coef_summary : `pandas.DataFrame`
        The filtered coefficient summary df filtered by the given conditions.
    """
    coef_summary_categorized = add_category_cols(coef_summary, pred_category)
    coef_summary_categorized["Pred_col"] = simplify_pred_cols(coef_summary_categorized["Pred_col"])
    # And True values will be aggregated with "or".
    # Any False values will be aggregated with "and".
    # This aligns with human sense.
    conditions_true = []
    conditions_false = []
    for k, v in locals().items():
        if "is_" in k and k != "is_interaction":
            if v is not None:
                if v:
                    conditions_true.append(f"{k} == {int(v)}")
                else:
                    conditions_false.append(f"{k} == {int(v)}")
    query_true = f"{' or '.join(conditions_true)}"
    query_false = f"{' and '.join(conditions_false)}"
    if query_true and query_false:
        query = f"({query_true}) and ({query_false})"
    elif query_true:
        query = f"({query_true})"
    elif query_false:
        query = f"({query_false})"
    else:
        query = ""
    # Deal with is_interaction
    if query:
        if is_interaction is False:
            # When is_interaction is False, remove interaction terms.
            # By default, the filters include interaction terms.
            query = f"({query}) and (is_interaction == {int(is_interaction)})"
        new_df = coef_summary_categorized.query(query)
    else:
        # In there is no query, is_interaction is used filter rows.
        if is_interaction is not None:
            query = f"is_interaction == {int(is_interaction)}"
            new_df = coef_summary_categorized.query(query)
        else:
            new_df = coef_summary_categorized
    return new_df.reset_index(drop=True)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen, Sayan Patra
"""Plotting functions in plotly."""

import warnings

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

from greykite.common import constants as cst
from greykite.common.features.timeseries_features import build_time_features_df
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import update_dictionary
from greykite.common.viz.colors_utils import get_color_palette


def plot_multivariate(
        df,
        x_col,
        y_col_style_dict="plotly",
        default_color="rgba(0, 145, 202, 1.0)",
        xlabel=None,
        ylabel=cst.VALUE_COL,
        title=None,
        showlegend=True):
    """Plots one or more lines against the same x-axis values.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Data frame with ``x_col`` and columns named by the keys in ``y_col_style_dict``.
    x_col: `str`
        Which column to plot on the x-axis.
    y_col_style_dict: `dict` [`str`, `dict` or None] or "plotly" or "auto" or "auto-fill", default "plotly"
        The column(s) to plot on the y-axis, and how to style them.

        If a dictionary:

            - key : `str`
                column name in ``df``
            - value : `dict` or None
                Optional styling options, passed as kwargs to `go.Scatter`.
                If None, uses the default: line labeled by the column name.
                See reference page for `plotly.graph_objs.Scatter` for options
                (e.g. color, mode, width/size, opacity).
                https://plotly.com/python/reference/#scatter.

        If a string, plots all columns in ``df`` besides ``x_col`` against ``x_col``:

            - "plotly": plot lines with default plotly styling
            - "auto": plot lines with color ``default_color``, sorted by value (ascending)
            - "auto-fill": plot lines with color ``default_color``, sorted by value (ascending), and fills between lines

    default_color: `str`, default "rgba(0, 145, 202, 1.0)" (blue)
        Default line color when ``y_col_style_dict`` is one of "auto", "auto-fill".
    xlabel : `str` or None, default None
        x-axis label. If None, default is ``x_col``.
    ylabel : `str` or None, default ``VALUE_COL``
        y-axis label
    title : `str` or None, default None
        Plot title. If None, default is based on axis labels.
    showlegend : `bool`, default True
        Whether to show the legend.

    Returns
    -------
    fig : `plotly.graph_objs.Figure`
        Interactive plotly graph of one or more columns
        in ``df`` against ``x_col``.

        See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
        return value for how to plot the figure and add customization.
    """

    if xlabel is None:
        xlabel = x_col
    if title is None and ylabel is not None:
        title = f"{ylabel} vs {xlabel}"

    auto_style = {"line": {"color": default_color}}
    if y_col_style_dict == "plotly":
        # Uses plotly default style
        y_col_style_dict = {col: None for col in df.columns if col != x_col}
    elif y_col_style_dict in ["auto", "auto-fill"]:
        # Columns ordered from low to high
        means = df.drop(columns=x_col).mean()
        column_order = list(means.sort_values().index)
        if y_col_style_dict == "auto":
            # Lines with color `default_color`
            y_col_style_dict = {col: auto_style for col in column_order}
        elif y_col_style_dict == "auto-fill":
            # Lines with color `default_color`, with fill between lines
            y_col_style_dict = {column_order[0]: auto_style}
            y_col_style_dict.update({
                col: {
                    "line": {"color": default_color},
                    "fill": "tonexty"
                } for col in column_order[1:]
            })

    data = []
    default_style = dict(mode="lines")
    for column, style_dict in y_col_style_dict.items():
        # By default, column name in ``df`` is used to label the line
        default_col_style = update_dictionary(default_style, overwrite_dict={"name": column})
        # User can overwrite any of the default values, or remove them by setting key value to None
        style_dict = update_dictionary(default_col_style, overwrite_dict=style_dict)
        line = go.Scatter(
            x=df[x_col],
            y=df[column],
            **style_dict)
        data.append(line)

    layout = go.Layout(
        xaxis=dict(title=xlabel),
        yaxis=dict(title=ylabel),
        title=title,
        showlegend=showlegend,
        legend={'traceorder': 'reversed'}  # Matches the order of ``y_col_style_dict`` (bottom to top)
    )
    fig = go.Figure(data=data, layout=layout)
    return fig


def plot_multivariate_grouped(
        df,
        x_col,
        y_col_style_dict,
        grouping_x_col,
        grouping_x_col_values,
        grouping_y_col_style_dict,
        colors=DEFAULT_PLOTLY_COLORS,
        xlabel=None,
        ylabel=cst.VALUE_COL,
        title=None,
        showlegend=True):
    """Plots multiple lines against the same x-axis values. The lines can
    partially share the x-axis values.

    See parameter descriptions for a running example.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Data frame with ``x_col`` and columns named by the keys in ``y_col_style_dict``,
        ``grouping_x_col``, ``grouping_y_col_style_dict``.

        For example::

            df = pd.DataFrame({
                time: [dt(2018, 1, 1),
                        dt(2018, 1, 2),
                        dt(2018, 1, 3)],
                "y1": [8.5, 2.0, 3.0],
                "y2": [1.4, 2.1, 3.4],
                "y3": [4.2, 3.1, 3.0],
                "y4": [0, 1, 2],
                "y5": [10, 9, 8],
                "group": [1, 2, 1],
            })
        This will be our running example.
    x_col: `str`
        Which column to plot on the x-axis.
        "time" in our example.
    y_col_style_dict: `dict` [`str`, `dict` or None]
        The column(s) to plot on the y-axis, and how to style them.
        These columns are plotted against the complete x-axis.

        - key : `str`
            column name in ``df``
        - value : `dict` or None
            Optional styling options, passed as kwargs to `go.Scatter`.
            If None, uses the default: line labeled by the column name.
            If line color is not given, it is added according to ``colors``.
            See reference page for `plotly.graph_objs.Scatter` for options
            (e.g. color, mode, width/size, opacity).
            https://plotly.com/python/reference/#scatter.

        For example::

            y_col_style_dict={
                "y1": {
                    "name": "y1_name",
                    "legendgroup": "one",
                    "mode": "markers",
                    "line": None  # Remove line params since we use mode="markers"
                },
                "y2": None,
            }

        The function will add a line color to "y1" and "y2" based on the ``colors`` parameter.
        It will also add a name to "y2", since none was given. The "name" of "y1" will be preserved.

        The output ``fig`` will have one line each for each of "y1" and "y2", each plot against
        the entire "time" column.
    grouping_x_col: `str`
        Which column to use to group columns in ``grouping_y_col_style_dict``.
        "group" in our example.
    grouping_x_col_values: `list` [`int`] or None
        Which values to use for grouping. If None, uses all the unique values in
        ``df`` [``grouping_x_col``].
        In our example, specifying ``grouping_x_col_values == [1, 2]`` would plot
        separate lines corresponding to ``group==1`` and ``group==2``.
    grouping_y_col_style_dict: `dict` [`str`, `dict` or None]
        The column(s) to plot on the y-axis, and how to style them.
        These columns are plotted against partial x-axis.
        For each ``grouping_x_col_values`` an element in this dictionary produces
        one line.

        - key : `str`
            column name in ``df``
        - value : `dict` or None
            Optional styling options, passed as kwargs to `go.Scatter`.
            If None, uses the default: line labeled by the ``grouping_x_col_values``,
            ``grouping_x_col`` and column name.
            If a name is given, it is augmented with the ``grouping_x_col_values``.
            If line color is not given, it is added according to ``colors``.
            All the lines sharing same ``grouping_x_col_values`` have the same color.
            See reference page for `plotly.graph_objs.Scatter` for options
            (e.g. color, mode, width/size, opacity).
            https://plotly.com/python/reference/#scatter.

        For example::

            grouping_y_col_style_dict={
                "y3": {
                    "line": {
                        "color": "blue"
                    }
                },
                "y4": {
                    "name": "y4_name",
                    "line": {
                        "width": 2,
                        "dash": "dot"
                    }
                },
                "y5": None,
            }

        The function will add a line color to "y4" and "y5" based on the ``colors`` parameter.
        The line color of "y3" will be "blue" as specified. We also preserve the given line
        properties of "y4".

    `   The function adds a name to "y3" and "y5", since none was given. The given "name" of "y4"
        will be augmented with ``grouping_x_col_values``.

        Each element of ``grouping_y_col_style_dict`` gets one line for each ``grouping_x_col_values``.
        In our example, there will be 2 lines corresponding to "y3", named "1_y3" and "2_y3".
        "1_y3" is plotted against "time = [dt(2018, 1, 1), dt(2018, 1, 3)]", corresponding to ``group==1``.
        "2_y3" is plotted against "time = [dt(2018, 1, 2)", corresponding to ``group==2``.
    colors: [`str`, `list` [`str`]], default ``DEFAULT_PLOTLY_COLORS``
        Which colors to use to build a color palette for plotting.
        This can be a list of RGB colors or a `str` from ``PLOTLY_SCALES``.
        Required number of colors equals sum of the length of ``y_col_style_dict``
        and length of ``grouping_x_col_values``.
        See `~greykite.common.viz.colors_utils.get_color_palette` for details.
    xlabel : `str` or None, default None
        x-axis label. If None, default is ``x_col``.
    ylabel : `str` or None, default ``VALUE_COL``
        y-axis label
    title : `str` or None, default None
        Plot title. If None, default is based on axis labels.
    showlegend : `bool`, default True
        Whether to show the legend.

    Returns
    -------
    fig : `plotly.graph_objs.Figure`
    Interactive plotly graph of one or more columns
    in ``df`` against ``x_col``.

    See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
    return value for how to plot the figure and add customization.
    """

    available_grouping_x_col_values = np.unique(df[grouping_x_col])
    if grouping_x_col_values is None:
        grouping_x_col_values = available_grouping_x_col_values
    else:
        missing_grouping_x_col_values = set(grouping_x_col_values) - set(available_grouping_x_col_values)
        if len(missing_grouping_x_col_values) > 0:
            raise ValueError(f"Following 'grouping_x_col_values' are missing in '{grouping_x_col}' column: "
                             f"{missing_grouping_x_col_values}")

    # Chooses the color palette
    n_color = len(y_col_style_dict) + len(grouping_x_col_values)
    color_palette = get_color_palette(num=n_color, colors=colors)

    # Updates colors for y_col_style_dict if it is not specified
    for color_num, (column, style_dict) in enumerate(y_col_style_dict.items()):
        if style_dict is None:
            style_dict = {}
        default_color = {"color": color_palette[color_num]}
        style_dict["line"] = update_dictionary(default_color, overwrite_dict=style_dict.get("line"))
        y_col_style_dict[column] = style_dict

    # Standardizes dataset for the next figure
    df_standardized = df.copy().drop_duplicates(subset=[x_col]).sort_values(by=x_col)

    # This figure plots the whole xaxis vs yaxis values
    fig = plot_multivariate(
        df=df_standardized,
        x_col=x_col,
        y_col_style_dict=y_col_style_dict,
        xlabel=xlabel,
        ylabel=ylabel,
        title=title,
        showlegend=showlegend)
    data = fig.data
    layout = fig.layout

    # These figures plot the sliced xaxis vs yaxis values
    for color_num, grouping_x_col_value in enumerate(grouping_x_col_values, len(y_col_style_dict)):
        default_color = {"color": color_palette[color_num]}

        sliced_y_col_style_dict = grouping_y_col_style_dict.copy()

        for column, style_dict in sliced_y_col_style_dict.items():
            # Updates colors if it is not specified
            if style_dict is None:
                style_dict = {}
            line_dict = update_dictionary(default_color, overwrite_dict=style_dict.get("line"))

            # Augments names with grouping_x_col_value
            name = style_dict.get("name")
            if name is None:
                updated_name = f"{grouping_x_col_value}_{grouping_x_col}_{column}"
            else:
                updated_name = f"{grouping_x_col_value}_{name}"

            overwrite_dict = {
                "name": updated_name,
                "line": line_dict
            }
            style_dict = update_dictionary(style_dict, overwrite_dict=overwrite_dict)
            sliced_y_col_style_dict[column] = style_dict

        df_sliced = df[df[grouping_x_col] == grouping_x_col_value]
        fig = plot_multivariate(
            df=df_sliced,
            x_col=x_col,
            y_col_style_dict=sliced_y_col_style_dict)
        data = data + fig.data

    fig = go.Figure(data=data, layout=layout)

    return fig


def plot_univariate(
        df,
        x_col,
        y_col,
        xlabel=None,
        ylabel=None,
        title=None,
        color="rgb(32, 149, 212)",  # light blue
        showlegend=True):
    """Simple plot of univariate timeseries.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Data frame with ``x_col`` and ``y_col``
    x_col: `str`
        x-axis column name, usually the time column
    y_col: `str`
        y-axis column name, the value the plot
    xlabel : `str` or None, default None
        x-axis label
    ylabel : `str` or None, default None
        y-axis label
    title : `str` or None, default None
        Plot title. If None, default is based on axis labels.
    color : `str`, default "rgb(32, 149, 212)" (light blue)
        Line color
    showlegend : `bool`, default True
        Whether to show the legend

    Returns
    -------
    fig : `plotly.graph_objs.Figure`
        Interactive plotly graph of the value against time.

        See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
        return value for how to plot the figure and add customization.

    See Also
    --------
    `~greykite.common.viz.timeseries_plotting.plot_multivariate`
        Provides more styling options. Also consider using plotly's `go.Scatter` and `go.Layout` directly.
    """
    # sets default x and y-axis names based on column names
    if xlabel is None:
        xlabel = x_col
    if ylabel is None:
        ylabel = y_col

    y_col_style_dict = {
        y_col: dict(
            name=y_col,
            mode="lines",
            line=dict(
                color=color
            ),
            opacity=0.8
        )
    }
    return plot_multivariate(
        df,
        x_col,
        y_col_style_dict,
        xlabel=xlabel,
        ylabel=ylabel,
        title=title,
        showlegend=showlegend,
    )


def plot_forecast_vs_actual(
        df,
        time_col=cst.TIME_COL,
        actual_col=cst.ACTUAL_COL,
        predicted_col=cst.PREDICTED_COL,
        predicted_lower_col=cst.PREDICTED_LOWER_COL,
        predicted_upper_col=cst.PREDICTED_UPPER_COL,
        xlabel=cst.TIME_COL,
        ylabel=cst.VALUE_COL,
        train_end_date=None,
        title=None,
        showlegend=True,
        actual_mode="lines+markers",
        actual_points_color="rgba(250, 43, 20, 0.7)",  # red
        actual_points_size=2.0,
        actual_color_opacity=1.0,
        forecast_curve_color="rgba(0, 90, 181, 0.7)",  # blue
        forecast_curve_dash="solid",
        ci_band_color="rgba(0, 90, 181, 0.15)",  # light blue
        ci_boundary_curve_color="rgba(0, 90, 181, 0.5)",  # light blue
        ci_boundary_curve_width=0.0,  # no line
        vertical_line_color="rgba(100, 100, 100, 0.9)",  # black color with opacity of 0.9
        vertical_line_width=1.0):
    """Plots forecast with prediction intervals, against actuals
    Adapted from plotly user guide:
    https://plot.ly/python/v3/continuous-error-bars/#basic-continuous-error-bars

    Parameters
    ----------
    df : `pandas.DataFrame`
        Timestamp, predicted, and actual values
    time_col : `str`, default `~greykite.common.constants.TIME_COL`
        Column in df with timestamp (x-axis)
    actual_col : `str`, default `~greykite.common.constants.ACTUAL_COL`
        Column in df with actual values
    predicted_col : `str`, default `~greykite.common.constants.PREDICTED_COL`
        Column in df with predicted values
    predicted_lower_col : `str` or None, default `~greykite.common.constants.PREDICTED_LOWER_COL`
        Column in df with predicted lower bound
    predicted_upper_col : `str` or None, default `~greykite.common.constants.PREDICTED_UPPER_COL`
        Column in df with predicted upper bound
    xlabel : `str`, default `~greykite.common.constants.TIME_COL`
        x-axis label.
    ylabel : `str`, default `~greykite.common.constants.VALUE_COL`
        y-axis label.
    train_end_date : `datetime.datetime` or None, default None
        Train end date.
        Must be a value in ``df[time_col]``.
    title : `str` or None, default None
        Plot title.
    showlegend : `bool`, default True
        Whether to show a plot legend.
    actual_mode : `str`, default "lines+markers"
        How to show the actuals.
        Options: ``markers``, ``lines``, ``lines+markers``
    actual_points_color : `str`, default "rgba(99, 114, 218, 1.0)"
        Color of actual line/marker.
    actual_points_size : `float`, default 2.0
        Size of actual markers.
        Only used if "markers" is in ``actual_mode``.
    actual_color_opacity : `float` or None, default 1.0
        Opacity of actual values points.
    forecast_curve_color : `str`, default "rgba(0, 145, 202, 1.0)"
        Color of forecasted values.
    forecast_curve_dash : `str`, default "solid"
        'dash' property of forecast ``scatter.line``.
        One of: ``['solid', 'dot', 'dash', 'longdash', 'dashdot', 'longdashdot']``
        or a string containing a dash length list in pixels or percentages
        (e.g. ``'5px 10px 2px 2px'``, ``'5, 10, 2, 2'``, ``'10% 20% 40%'``)
    ci_band_color : `str`, default "rgba(0, 145, 202, 0.15)"
        Fill color of the prediction bands.
    ci_boundary_curve_color : `str`, default "rgba(0, 145, 202, 0.15)"
        Color of the prediction upper/lower lines.
    ci_boundary_curve_width : `float`, default 0.0
        Width of the prediction upper/lower lines.
        default 0.0 (hidden)
    vertical_line_color : `str`, default "rgba(100, 100, 100, 0.9)"
        Color of the vertical line indicating train end date.
        Default is black with opacity of 0.9.
    vertical_line_width : `float`, default 1.0
        width of the vertical line indicating train end date

    Returns
    -------
    fig : `plotly.graph_objs.Figure`
        Plotly figure of forecast against actuals, with prediction
        intervals if available.

        Can show, convert to HTML, update::

            # show figure (saved to HTML)
            from plotly.offline import plot
            plot(fig)
            # show figure in notebook
            from plotly.offline import init_notebook_mode, iplot
            init_notebook_mode(connected=True)
            iplot(fig)

            # get HTML string, write to file
            fig.to_html(include_plotlyjs=False, full_html=True)
            fig.write_html("figure.html", include_plotlyjs=False, full_html=True)

            # customize layout (https://plot.ly/python/v3/user-guide/)
            update_layout = dict(
                yaxis=dict(title="new ylabel"),
                title_text="new title",
                title_font_size=30)
            fig.update(layout=update_layout)
    """
    if title is None:
        title = "Forecast vs Actual"
    if train_end_date is not None and not all(pd.Series(train_end_date).isin(df[time_col])):
        raise Exception(
            f"train_end_date {train_end_date} is not found in df['{time_col}']")

    fill_dict = {
        "mode": "lines",
        "fillcolor": ci_band_color,
        "fill": "tonexty"
    }
    data = []
    if predicted_lower_col is not None:
        lower_bound = go.Scatter(
            name="Lower Bound",
            x=df[time_col],
            y=df[predicted_lower_col],
            mode="lines",
            line=dict(
                width=ci_boundary_curve_width,
                color=ci_boundary_curve_color),
            legendgroup="interval"  # show/hide with the upper bound
        )
        data.append(lower_bound)

    # plotly fills between current and previous element in `data`.
    # Only fill if lower bound exists.
    forecast_fill_dict = fill_dict if predicted_lower_col is not None else {}
    if predicted_upper_col is not None:
        upper_bound = go.Scatter(
            name="Upper Bound",
            x=df[time_col],
            y=df[predicted_upper_col],
            line=dict(
                width=ci_boundary_curve_width,
                color=ci_boundary_curve_color),
            legendgroup="interval",  # show/hide with the lower bound
            **forecast_fill_dict)
        data.append(upper_bound)

    # If `predicted_lower_col` and `predicted_upper_col`, then the full range
    # has been filled in. If only one of them, then fill in between that line
    # and forecast.
    actual_params = {}
    if "lines" in actual_mode:
        actual_params.update(line=dict(color=actual_points_color))
    if "markers" in actual_mode:
        actual_params.update(marker=dict(color=actual_points_color, size=actual_points_size))
    actual = go.Scatter(
        name="Actual",
        x=df[time_col],
        y=df[actual_col],
        mode=actual_mode,
        opacity=actual_color_opacity,
        **actual_params
    )
    data.append(actual)

    forecast_fill_dict = fill_dict if (predicted_lower_col is None) != (predicted_upper_col is None) else {}
    forecast = go.Scatter(
        name="Forecast",
        x=df[time_col],
        y=df[predicted_col],
        line=dict(
            color=forecast_curve_color,
            dash=forecast_curve_dash),
        **forecast_fill_dict)
    data.append(forecast)

    layout = go.Layout(
        xaxis=dict(title=xlabel),
        yaxis=dict(title=ylabel),
        title=title,
        showlegend=showlegend,
        # legend order from top to bottom: Actual, Forecast, Upper Bound, Lower Bound
        legend={'traceorder': 'reversed'}
    )
    fig = go.Figure(data=data, layout=layout)
    fig.update()

    # adds a vertical line to separate training and testing phases
    if train_end_date is not None:
        new_layout = dict(
            # add vertical line
            shapes=[dict(
                type="line",
                xref="x",
                yref="paper",  # y-reference is assigned to the plot paper [0,1]
                x0=train_end_date,
                y0=0,
                x1=train_end_date,
                y1=1,
                line=dict(
                    color=vertical_line_color,
                    width=vertical_line_width)
            )],
            # add text annotation
            annotations=[dict(
                xref="x",
                x=train_end_date,
                yref="paper",
                y=.97,
                text="Train End Date",
                showarrow=True,
                arrowhead=0,
                ax=-60,
                ay=0
            )]
        )
        fig.update(layout=new_layout)
    return fig


def split_range_into_groups(
        n,
        group_size,
        which_group_complete="last"):
    """Partitions `n` elements into adjacent groups,
        each with `group_size` elements
        Group number starts from 0 and increments upward
        Can be used to generate groups for sliding window aggregation.
    :param n: int
        number of elemnts to split into groups
    :param group_size: int
        number of elements per group
    :param which_group_complete: str
        If n % group_size > 0, one group will have fewer than `group_size` elements
        if "first", the first group is full if possible, and last group may be incomplete
        if "last", (default) the last group is full if possible,
        and first group may be incomplete
    :return: np.array of length n
        values correspond to the element's group number

    Examples:
    >>> split_range_into_groups(10, 1, "last")
    array([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> split_range_into_groups(10, 2, "last")
    array([0., 0., 1., 1., 2., 2., 3., 3., 4., 4.])
    >>> split_range_into_groups(10, 3, "last")
    array([0., 1., 1., 1., 2., 2., 2., 3., 3., 3.])
    >>> split_range_into_groups(10, 4, "last")
    array([0., 0., 1., 1., 1., 1., 2., 2., 2., 2.])
    >>> split_range_into_groups(10, 4, "first")
    array([0., 0., 0., 0., 1., 1., 1., 1., 2., 2.])
    >>> split_range_into_groups(10, 5, "last")
    array([0., 0., 0., 0., 0., 1., 1., 1., 1., 1.])
    >>> split_range_into_groups(10, 6, "last")
    array([0., 0., 0., 0., 1., 1., 1., 1., 1., 1.])
    >>> split_range_into_groups(10, 10, "last")
    array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
    >>> split_range_into_groups(10, 12, "last")
    array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
    """
    if which_group_complete.lower() == "first":
        offset = 0
    else:
        offset = group_size - n % group_size
        offset = offset % group_size  # sets offset to 0 if n % group_size == 0
    return np.floor(np.arange(offset, n + offset) / group_size)


def add_groupby_column(
        df,
        time_col,
        groupby_time_feature=None,
        groupby_sliding_window_size=None,
        groupby_custom_column=None):
    """Extracts a column to group by from ``df``.

    Exactly one of ``groupby_time_feature``, ``groupby_sliding_window_size``,
    `groupby_custom_column` must be provided.

    Parameters
    ----------
    df : 'pandas.DataFrame`
        Contains the univariate time series / forecast
    time_col : `str`
        The name of the time column of the univariate time series / forecast
    groupby_time_feature : `str` or None, optional
        If provided, groups by a column generated by
        `~greykite.common.features.timeseries_features.build_time_features_df`.
        See that function for valid values.
    groupby_sliding_window_size : `int` or None, optional
        If provided, sequentially partitions data into groups of size
        ``groupby_sliding_window_size``.
    groupby_custom_column : `pandas.Series` or None, optional
        If provided, groups by this column value.
        Should be same length as the ``df``.

    Returns
    -------
    result : `dict`
        Dictionary with two items:

        * ``"df"`` : `pandas.DataFrame`
            ``df`` with a grouping column added.
            The column can be used to group rows together.

        * ``"groupby_col"`` : `str`
            The name of the groupby column added to ``df``.
            The column name depends on the grouping method:

                - ``groupby_time_feature`` for ``groupby_time_feature``
                - ``{cst.TIME_COL}_downsample`` for ``groupby_sliding_window_size``
                - ``groupby_custom_column.name`` for ``groupby_custom_column``.
    """
    # Resets index to support indexing in groupby_sliding_window_size
    df = df.copy()
    dt = pd.Series(df[time_col].values)
    # Determines the groups
    is_groupby_time_feature = 1 if groupby_time_feature is not None else 0
    is_groupby_sliding_window_size = 1 if groupby_sliding_window_size is not None else 0
    is_groupby_custom_column = 1 if groupby_custom_column is not None else 0
    if is_groupby_time_feature + is_groupby_sliding_window_size + is_groupby_custom_column != 1:
        raise ValueError(
            "Exactly one of (groupby_time_feature, groupby_rolling_window_size, groupby_custom_column)"
            "must be specified")
    groups = None
    if is_groupby_time_feature == 1:
        # Group by a value derived from the time column
        time_features = build_time_features_df(dt, conti_year_origin=min(dt).year)
        groups = time_features[groupby_time_feature]
        groups.name = groupby_time_feature
    elif is_groupby_sliding_window_size == 1:
        # Group by sliding window for evaluation over time
        index_dates = split_range_into_groups(
            n=df.shape[0],
            group_size=groupby_sliding_window_size,
            which_group_complete="last")  # ensures the last group is complete (first group may be partial)
        groups = dt[index_dates * groupby_sliding_window_size]  # uses first date in each group as grouping value
        groups.name = f"{time_col}_downsample"
    elif is_groupby_custom_column == 1:
        # Group by custom column
        groups = groupby_custom_column

    groups_col_name = groups.name if groups.name is not None else "groups"
    df[groups_col_name] = groups.values
    if df.index.name in df.columns:
        # Removes ambiguity in case the index name is the same as the newly added column,
        # (or an existing column).
        df.index.name = None
    return {
        "df": df,
        "groupby_col": groups_col_name
    }


def grouping_evaluation(
        df,
        groupby_col,
        grouping_func,
        grouping_func_name):
    """Groups ``df`` and evaluates a function on each group.
    The function takes a `pandas.DataFrame` and returns a scalar.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input data. For example, univariate time series, or forecast result.
        Contains ``groupby_col`` and columns to apply ``grouping_func`` on.
    groupby_col : `str`
        Column name in ``df`` to group by.
    grouping_func : `callable`
        Function that is applied to each group via `pandas.groupBy.apply`.
        Signature (grp: `pandas.DataFrame`) -> aggregated value: `float`.
    grouping_func_name : `str`
        What to call the output column generated by ``grouping_func``.

    Returns
    -------
    grouped_df : `pandas.DataFrame`
        Dataframe with ``grouping_func`` evaluated on each level of ``df[groupby_col]``.
        Contains two columns:

            - ``groupby_col``: The groupby value
            - ``grouping_func_name``: The output of ``grouping_func`` on the group
    """
    grouped_df = (df
                  .groupby(groupby_col)
                  .apply(grouping_func)
                  .reset_index()
                  .rename({0: grouping_func_name}, axis=1))

    return grouped_df


def flexible_grouping_evaluation(
        df,
        map_func_dict=None,
        groupby_col=None,
        agg_kwargs=None,
        extend_col_names=True,
        unpack_list=True,
        list_names_dict=None):
    """Flexible aggregation. Generates additional columns for evaluation via
    ``map_func_dict``, groups by ``groupby_col``, then aggregates according
    to ``agg_kwargs``.

    This function calls `pandas.DataFrame.apply` and
    `pandas.core.groupby.DataFrameGroupBy.agg` internally.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame to transform / aggregate
    map_func_dict : `dict` [`str`, `callable`] or None, default None
        Row-wise transformation functions to create new columns.
        If None, no new columns are added.

        key: new column name
        value: row-wise function to apply to ``df`` to generate the column value.
               Signature (row: `pandas.DataFrame`) -> transformed value: `float`.

        For example::

            map_func_dict = {
                "residual": lambda row: row["predicted"] - row["actual"],
                "squared_error": lambda row: (row["predicted"] - row["actual"])**2
            }

    groupby_col : `str` or None, default None
        Which column to group by.
        Can be in ``df`` or generated by ``map_func_dict``.
        If None, no grouping or aggregation is done.
    agg_kwargs : `dict` or None, default None
        Passed as keyword args to `pandas.core.groupby.DataFrameGroupBy.aggregate` after creating
        new columns and grouping by ``groupby_col``. Must be provided if ``groupby_col is not None``.
        To fully customize output column names, pass a dictionary as shown below.

        For example::

            # Example 1, named aggregation to explicitly name output columns.
            # Assume ``df`` contains ``abs_percent_err``, ``abs_err`` columns.
            # Output columns are "MedAPE", "MAPE", "MAE", etc. in a single level index.
            from functools import partial
            agg_kwargs = {
                # output column name: (column to aggregate, aggregation function)
                "MedAPE": pd.NamedAgg(column="abs_percent_err", aggfunc=np.nanmedian),
                "MAPE": pd.NamedAgg(column="abs_percent_err", aggfunc=np.nanmean),
                "MAE": pd.NamedAgg(column="abs_err", aggfunc=np.nanmean),
                "q95_abs_err": pd.NamedAgg(column="abs_err", aggfunc=partial(np.nanquantile, q=0.95)),
                "q05_abs_err": pd.NamedAgg(column="abs_err", aggfunc=partial(np.nanquantile, q=0.05)),
            }

            # Example 2, multi-level aggregation using `func` parameter
            # to `pandas.core.groupby.DataFrameGroupBy.aggregate`.
            # Assume ``df`` contains ``y1``, ``y2`` columns.
            agg_kwargs = {
                "func": {
                    "y1": [np.nanmedian, np.nanmean],
                    "y2": [np.nanmedian, np.nanmax],
                }
            }
            # `extend_col_names` controls the output column names
            extend_col_names = True  # output columns are "y1_nanmean", "y1_nanmedian", "y2_nanmean", "y2_nanmax"
            extend_col_names = False  # output columns are "nanmean", "nanmedian", "nanmean", "nanmax"

    extend_col_names : `bool` or None, default True
        How to flatten index after aggregation.
        In some cases, the column index after aggregation is a multi-index.
        This parameter controls how to flatten an index with 2 levels to 1 level.

            - If None, the index is not flattened.
            - If True, column name is a composite: ``{index0}_{index1}``
              Use this option if index1 is not unique.
            - If False, column name is simply ``{index1}``

        Ignored if the ColumnIndex after aggregation has only one level (e.g.
        if named aggregation is used in ``agg_kwargs``).

    unpack_list : `bool`, default True
        Whether to unpack (flatten) columns that contain list/tuple after aggregation,
        to create one column per element of the list/tuple.
        If True, ``list_names_dict`` can be used to rename the unpacked columns.

    list_names_dict : `dict` [`str`, `list` [`str`]] or None, default None
        If ``unpack_list`` is True, this dictionary can optionally be
        used to rename the unpacked columns.

            - Key = column name after aggregation, before upacking.
              E.g. ``{index0}_{index1}`` or ``{index1}`` depending on ``extend_col_names``.
            - Value = list of names to use for the unpacked columns. Length must match
              the length of the lists contained in the column.

        If a particular list/tuple column is not found in this dictionary, appends
        0, 1, 2, ..., n-1 to the original column name, where n = list length.

        For example, if the column contains a tuple of length 4 corresponding to
        quantiles 0.1, 0.25, 0.75, 0.9, then the following would be appropriate::

            aggfunc = lambda grp: partial(np.nanquantile, q=[0.1, 0.25, 0.75, 0.9])(grp).tolist()
            agg_kwargs = {
                "value_Q": pd.NamedAgg(column="value", aggfunc=aggfunc)
            }
            list_names_dict = {
                # the key is the name of the unpacked column
                "value_Q" : ["Q0.10", "Q0.25", "Q0.75", "Q0.90"]
            }
            # Output columns are "Q0.10", "Q0.25", "Q0.75", "Q0.90"

            # In this example, if list_names_dict=None, the default output column names
            # would be: "value_Q0", "value_Q1", "value_Q2", "value_Q3"

    Returns
    -------
    df_transformed : `pandas.DataFrame`
        df after transformation and optional aggregation.

        If ``groupby_col`` is None, returns ``df`` with additional columns as the keys in ``map_func_dict``.
        Otherwise, ``df`` is grouped by ``groupby_col`` and this becomes the index. Columns
        are determined by ``agg_kwargs`` and ``extend_col_names``.
    """
    if groupby_col and not agg_kwargs:
        raise ValueError("Must specify `agg_kwargs` if grouping is requested via `groupby_col`.")
    if agg_kwargs and not groupby_col:
        log_message(f"`agg_kwargs` is ignored because `groupby_col` is None. "
                    f"Specify `groupby_col` to allow aggregation.", LoggingLevelEnum.WARNING)

    df = df.copy()
    if map_func_dict is not None:
        for col_name, func in map_func_dict.items():
            df[col_name] = df.apply(func, axis=1)

    if groupby_col is not None:
        groups = df.groupby(groupby_col)
        with warnings.catch_warnings():
            # Ignores pandas FutureWarning. Use NamedAgg in pandas 0.25.+
            warnings.filterwarnings(
                "ignore",
                message="using a dict with renaming is deprecated",
                category=FutureWarning)
            df_transformed = groups.agg(**agg_kwargs)
        if extend_col_names is not None and df_transformed.columns.nlevels > 1:
            # Flattens multi-level column index
            if extend_col_names:
                # By concatenating names
                df_transformed.columns = ["_".join(col).strip("_") for col in df_transformed.columns]
            else:
                # By using level 1 names
                df_transformed.columns = list(df_transformed.columns.get_level_values(1))
                if np.any(df_transformed.columns.duplicated()):
                    warnings.warn("Column names are not unique. Use `extend_col_names=True` "
                                  "to uniquely identify every column.")
    else:
        # No grouping is requested
        df_transformed = df

    if unpack_list and df_transformed.shape[0] > 0:
        # Identifies the columns that contain list elements
        which_list_cols = df_transformed.iloc[0].apply(lambda x: isinstance(x, (list, tuple)))
        list_cols = list(which_list_cols[which_list_cols].index)
        for col in list_cols:
            if isinstance(df_transformed[col], pd.DataFrame):
                warnings.warn(f"Skipping list unpacking for `{col}`. There are multiple columns "
                              f"with this name. Make sure column names are unique to enable unpacking.")
                continue
            # Unpacks the column, creating one column for each list entry
            list_df = pd.DataFrame(df_transformed[col].to_list())
            n_cols = list_df.shape[1]
            # Adds column names
            if list_names_dict is not None and col in list_names_dict:
                found_length = len(list_names_dict[col])
                if found_length != n_cols:
                    raise ValueError(
                        f"list_names_dict['{col}'] has length {found_length}, "
                        f"but there are {n_cols} columns to name. Example row(s):\n"
                        f"{list_df.head(2)}")
                list_df.columns = [f"{list_names_dict.get(col)[i]}" for i in range(n_cols)]
            else:
                list_df.columns = [f"{col}{i}" for i in range(n_cols)]
            # replaces original column with new ones
            list_df.index = df_transformed.index
            del df_transformed[col]
            df_transformed = pd.concat([df_transformed, list_df], axis=1)

        if list_names_dict:
            unused_names = sorted(list(set(list_names_dict.keys()) - set(list_cols)))
            if len(unused_names) > 0:
                warnings.warn("These names from `list_names_dict` are not used, because the "
                              "column (key) is not found in the dataframe after aggregation:\n"
                              f"{unused_names}.\nAvailable columns are:\n"
                              f"{list_cols}.")

    return df_transformed



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Cross validator for time series cross validation,
compatible with sklearn.
"""

import math
import random
import warnings

import numpy as np
from sklearn.model_selection import BaseCrossValidator

from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.common.python_utils import get_integer


class RollingTimeSeriesSplit(BaseCrossValidator):
    """Flexible splitter for time-series cross validation and rolling window evaluation.
    Suitable for use in GridSearchCV.

    Attributes
    ----------
    min_splits : int
        Guaranteed min number of splits. This is always set to 1. If provided configuration results in 0 splits,
        the cross validator will yield a default split.

    __starting_test_index : int
        Test end index of the first CV split. Actual offset = __starting_test_index + _get_offset(X), for a particular
        dataset X.
        Cross validator ensures the last test split contains the last observation in X.

    Examples
    --------
    >>> from greykite.sklearn.cross_validation import RollingTimeSeriesSplit
    >>> X = np.random.rand(20, 4)
    >>> tscv = RollingTimeSeriesSplit(forecast_horizon=3, max_splits=4)
    >>> tscv.get_n_splits(X=X)
    4
    >>> for train, test in tscv.split(X=X):
    ...     print(train, test)
    [2 3 4 5 6 7] [ 8  9 10]
    [ 5  6  7  8  9 10] [11 12 13]
    [ 8  9 10 11 12 13] [14 15 16]
    [11 12 13 14 15 16] [17 18 19]
    >>> X = np.random.rand(20, 4)
    >>> tscv = RollingTimeSeriesSplit(forecast_horizon=2,
    ...                               min_train_periods=4,
    ...                               expanding_window=True,
    ...                               periods_between_splits=4,
    ...                               periods_between_train_test=2,
    ...                               max_splits=None)
    >>> tscv.get_n_splits(X=X)
    4
    >>> for train, test in tscv.split(X=X):
    ...     print(train, test)
    [0 1 2 3] [6 7]
    [0 1 2 3 4 5 6 7] [10 11]
    [ 0  1  2  3  4  5  6  7  8  9 10 11] [14 15]
    [ 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15] [18 19]
    >>> X = np.random.rand(5, 4)  # default split if there is not enough data
    >>> for train, test in tscv.split(X=X):
    ...     print(train, test)
    [0 1 2 3] [4]
    """
    def __init__(
            self,
            forecast_horizon,
            min_train_periods=None,
            expanding_window=False,
            use_most_recent_splits=False,
            periods_between_splits=None,
            periods_between_train_test=0,
            max_splits=3):
        """Initializes attributes of RollingTimeSeriesSplit

        Parameters
        ----------
        forecast_horizon : `int`
            How many periods in each CV test set

        min_train_periods : `int` or None, optional
            Minimum number of periods for training.
            If ``expanding_window`` is False, every training period has this size.

        expanding_window : `bool`, default False
            If True, training window for each CV split is fixed to the first available date.
            Otherwise, train start date is sliding, determined by ``min_train_periods``.

        use_most_recent_splits: `bool`, default False
            If True, splits from the end of the dataset are used.
            Else a sampling strategy is applied. Check
            `~greykite.sklearn.cross_validation.RollingTimeSeriesSplit._sample_splits`
            for details.

        periods_between_splits : `int` or None
            Number of periods to slide the test window

        periods_between_train_test : `int`
            Number of periods gap between train and test within a CV split

        max_splits : `int` or None
            Maximum number of CV splits. Given the above configuration, samples up to max_splits train/test splits,
            preferring splits toward the end of available data. If None, uses all splits.
        """
        super().__init__()
        self.forecast_horizon = get_integer(forecast_horizon, name="forecast_horizon", min_value=1)

        # by default, use at least twice the forecast horizon for training
        self.min_train_periods = get_integer(min_train_periods, name="min_train_periods",
                                             min_value=1, default_value=2 * self.forecast_horizon)

        # by default, use fixed size training window
        self.expanding_window = expanding_window

        # by default, does not force most recent splits
        self.use_most_recent_splits = use_most_recent_splits

        # by default, use non-overlapping test sets
        self.periods_between_splits = get_integer(periods_between_splits, name="periods_between_splits",
                                                  min_value=1, default_value=self.forecast_horizon)

        # by default, use test set immediately following train set
        self.periods_between_train_test = get_integer(periods_between_train_test, name="periods_between_train_test",
                                                      min_value=0, default_value=0)

        if self.min_train_periods < 2 * self.forecast_horizon:
            warnings.warn(f"`min_train_periods` is too small for your `forecast_horizon`. Should be at least"
                          f" {forecast_horizon*2}=2*`forecast_horizon`.")

        self.max_splits = max_splits
        self.min_splits = 1  # CV ensures there is always at least one split
        # test end index for the first CV split, before applying offset to ensure last data point in X is used
        self.__starting_test_index = (self.forecast_horizon
                                      + self.min_train_periods
                                      + self.periods_between_train_test
                                      - 1)

    def split(self, X, y=None, groups=None):
        """Generates indices to split data into training and test CV folds according to rolling
          window time series cross validation

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
            Must have `shape` method.

        y : array-like, shape (n_samples,), optional
            The target variable for supervised learning problems. Always ignored, exists for compatibility.

        groups : array-like, with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set. Always ignored, exists for compatibility.

        Yields
        ------
        train : `numpy.array`
            The training set indices for that split.

        test : `numpy.array`
            The testing set indices for that split.
        """
        num_samples = X.shape[0]
        indices = np.arange(num_samples)

        n_splits_without_capping = self.get_n_splits_without_capping(X=X)
        n_splits = self.get_n_splits(X=X)
        if n_splits_without_capping == 0:
            warnings.warn("There are no CV splits under the requested settings. Decrease `forecast_horizon` and/or"
                          " `min_train_periods`. Using default 90/10 CV split")
        elif n_splits == 1:
            warnings.warn("There is only one CV split")
        elif n_splits >= 10:
            warnings.warn(f"There is a high number of CV splits ({n_splits}). If training is slow, increase "
                          f"`periods_between_splits` or `min_train_periods`, or decrease `max_splits`")

        log_message(f"There are {n_splits} CV splits.", LoggingLevelEnum.INFO)

        if n_splits_without_capping == 0:  # uses default split
            default_split_ratio = 0.9
            train_samples = int(round(num_samples * default_split_ratio))
            yield indices[:train_samples], indices[train_samples:]
        else:  # determines which splits to keep so that up to max_splits are returned
            splits_to_keep = self._sample_splits(n_splits_without_capping)

            last_index = num_samples - 1
            test_end_index = self.__starting_test_index + self._get_offset(X=X)
            current_split_index = 0
            while test_end_index <= last_index:
                test_start_index = test_end_index - self.forecast_horizon + 1
                train_end_index = test_start_index - self.periods_between_train_test - 1
                train_start_index = 0 if self.expanding_window else train_end_index - self.min_train_periods + 1
                assert train_start_index >= 0  # guaranteed by n_splits > 0

                if current_split_index in splits_to_keep:
                    log_message(f"CV split: Train {train_start_index} to {train_end_index}. "
                                f"Test {test_start_index} to {test_end_index}.", LoggingLevelEnum.DEBUG)
                    yield indices[train_start_index:train_end_index + 1], indices[test_start_index:test_end_index + 1]

                test_end_index += self.periods_between_splits
                current_split_index += 1

    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations yielded by the cross-validator

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data to split

        y : object
            Always ignored, exists for compatibility.

        groups : object
            Always ignored, exists for compatibility.

        Returns
        -------
        n_splits : int
            The number of splitting iterations yielded by the cross-validator.
        """
        num_splits = self.get_n_splits_without_capping(X=X)
        if self.max_splits is not None and num_splits > self.max_splits:
            num_splits = self.max_splits  # num_splits is set to max limit
        if num_splits == 0:
            num_splits = self.min_splits  # not enough observations to create split, uses default
        return num_splits

    def get_n_splits_without_capping(self, X=None):
        """Returns the number of splitting iterations in the cross-validator as configured, ignoring
            self.max_splits and self.min_splits

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data to split

        Returns
        -------
        n_splits : int
            The number of splitting iterations in the cross-validator as configured, ignoring
            self.max_splits and self.min_splits
        """
        last_index = X.shape[0] - 1
        starting_index = self.__starting_test_index + self._get_offset(X=X)
        if starting_index > last_index:
            return 0
        return math.ceil((last_index - starting_index + 1) / self.periods_between_splits)

    def _get_offset(self, X=None):
        """Returns an offset to add to test set indices when creating CV splits
        CV splits are shifted so that the last test observation is the last point in X.
        This shift does not affect the total number of splits.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data to split

        Returns
        -------
        offset : int
            The number of observations to ignore at the beginning of X when creating CV splits
        """
        last_index = X.shape[0] - 1
        starting_index = self.__starting_test_index
        if starting_index > last_index:
            return 0
        return (last_index - starting_index) % self.periods_between_splits

    def _sample_splits(self, num_splits, seed=48912):
        """Samples up to ``max_splits`` items from list(range(`num_splits`)).

        If ``use_most_recent_splits`` is True, highest split indices up to ``max_splits``
        are retained. Otherwise, the following sampling scheme is implemented:

            - takes the last 2 splits
            - samples from the rest uniformly at random

        Parameters
        ----------
        num_splits : `int`
            Number of splits before sampling.

        seed : `int`
            Seed for random sampling.

        Returns
        -------
        n_splits : `list`
            Indices of splits to keep (subset of `list(range(num_splits))`).
        """
        split_indices = list(range(num_splits))
        if self.max_splits is not None and num_splits > self.max_splits:
            if self.use_most_recent_splits:
                # keep indices from the end up to max_splits
                keep_split_indices = split_indices[-self.max_splits:]
            else:
                # applies sampling scheme to take up to max_splits
                keep_split_indices = []
                if self.max_splits > 0:  # first takes the last split
                    keep_split_indices.append(split_indices[-1])
                if self.max_splits > 1:  # then takes the second to last split
                    keep_split_indices.append(split_indices[-2])
                if self.max_splits > 2:  # then randomly samples the remaining splits
                    random.seed(seed)
                    keep_split_indices += random.sample(split_indices[:-2], self.max_splits - 2)
            split_indices = keep_split_indices
        return split_indices

    def _iter_test_indices(self, X=None, y=None, groups=None):
        """Class directly implements `split` instead of providing this function"""
        raise NotImplementedError



#!/usr/bin/env python

"""The setup script."""

import sys
from setuptools import setup, find_packages


with open("README_PYPI.rst") as readme_file:
    LONG_DESC = readme_file.read()


requirements = [
    "Cython==0.29.23",
    "cvxpy==1.1.12",
    "fbprophet==0.5",
    "holidays==0.9.10",  # 0.10.2,
    "ipykernel==4.8.2",
    "ipython==7.1.1",
    "ipywidgets==7.2.1",
    "jupyter==1.0.0",
    "jupyter-client==6.1.5",
    "jupyter-console==6.",  # used version 6 to avoid conflict with ipython version
    "jupyter-core==4.7.1",
    "matplotlib==3.4.1",
    "nbformat==5.1.3",
    "notebook==5.4.1",
    "numpy==1.20.2",
    "osqp==0.6.1",
    "overrides==2.8.0",
    "pandas==1.1.3",
    "patsy==0.5.1",
    "Pillow==8.0.1",
    "plotly==3.10.0",
    "pystan==2.18.0.0",
    "pyzmq==22.0.3",
    "scipy==1.5.4",
    "seaborn==0.9.0",
    "six==1.15.0",
    "scikit-learn==0.24.1",
    "Sphinx==3.2.1",
    "sphinx-gallery==0.6.1",
    "sphinx-rtd-theme==0.4.2",
    "statsmodels==0.12.2",
    "testfixtures==6.14.2",
    "tornado==5.1.1",
    "tqdm==4.52.0"]

# Here we change some dependencies versions according to python version
# to minimize the chance of install failures.
if sys.version_info < (3, 7):
    requirements = ["numpy==1.19.1" if i=="numpy==1.20.2" else i for i in requirements]

setup_requirements = ["pytest-runner", ]

test_requirements = ["pytest>=3", ]

setup(
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    description="A python package for flexible forecasting",
    long_description=LONG_DESC,
    entry_points={
        "console_scripts": [
            "greykite=greykite.cli:main",
        ],
    },
    install_requires=requirements,
    license="BSD-2-CLAUSE",
    include_package_data=True,
    keywords="greykite",
    name="greykite",
    author="R. Hosseini, A. Chen, K. Yang, S. Patra, R. Arora",
    author_email="reza1317@gmail.com",
    packages=find_packages(include=['greykite', 'greykite.*']),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/linkedin/greykite",
    version="0.1.0",
    zip_safe=False,
)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen

import warnings

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError


class DropDegenerateTransformer(BaseEstimator, TransformerMixin):
    """Removes degenerate (constant) columns.

    Parameters
    ----------
    drop_degenerate : `bool`, default False
        Whether to drop degenerate columns.

    Attributes
    ----------
    drop_cols : `list` [`str`] or None
        Degenerate columns to drop
    keep_cols : `list` [`str`] or None
        Columns to keep
    """

    def __init__(self, drop_degenerate=False):
        # sets params without modification to ensure get_params() works in grid search
        self.drop_degenerate = drop_degenerate

        self.drop_cols = None
        self.keep_cols = None

    def fit(self, X, y=None):
        """Identifies the degenerate columns, and sets ``self.keep_cols``
        and ``self.drop_cols``.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Training input data. e.g. each column is a timeseries.
            Columns are expected to be numeric.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        assert isinstance(X, pd.DataFrame)
        if self.drop_degenerate:
            self.keep_cols = list(X.loc[:, (X != X.iloc[0]).any()].columns)
            self.drop_cols = [col for col in X.columns if col not in self.keep_cols]
        else:
            self.keep_cols = list(X.columns)
            self.drop_cols = []

        if self.drop_cols:
            warnings.warn(f"Columns {self.drop_cols} are degenerate (constant value), "
                          f"and will not be used in the forecast.", RuntimeWarning)
        return self

    def transform(self, X):
        """Normalizes data using the specified scaling method.

        Parameters
        ----------
        X : `pandas.DataFrame`
            Data to transform. e.g. each column is a timeseries.
            Columns are expected to be numeric.

        Returns
        -------
        X_subset : `pandas.DataFrame`
            Selected columns of X. Keeps columns that were not
            degenerate on the training data.
        """
        if self.keep_cols is None:
            raise NotFittedError(
                "This instance is not fitted yet. Call 'fit' with appropriate arguments "
                "before calling 'transform'.")
        return X[self.keep_cols]
